"""
=======================================================================================
 * Compatible with TensorFlow V1.0 python hipster.
=======================================================================================
 hipster functionality for tensor flow.  This hipster provides neural network functionality for

---------------------------------------------------------------------------------------
 * class definitions
---------------------------------------------------------------------------------------
     Colours                   - Defines special characters for coloured outputs
     MLPTrainingOptions        - Options required to set up training for an MLP/DNN
     CNNTrainingOptions        - Options required to set up training for a CNN
     Fisher                    - Fisher discriminant computation class
     ROC                       - Class for caclulating ROC curves

---------------------------------------------------------------------------------------
 * General NN operations
---------------------------------------------------------------------------------------
     CountLabels               - Given a label tensor, compute the number of unique
                                 values and prepare a tensor for use with training.
     PrepareWeights            - Given an array of example weights, compute the sum of
                                 weights and prepare a tensor for use with training.
     WeightVariable            - Weights initialised with a truncated normal
                                 distribution
     BiasVariable              - Biases intiialised with a small positive offset
     NetworkActivationFunction - Interface to the tf.nn activation function types;
                                 implemented for use with MLPs and CNNs.

---------------------------------------------------------------------------------------
 * Convolutional Neural Networks
---------------------------------------------------------------------------------------
     BuildCNN                  - Interface function to instantiate a CNN model of a
                                 given type
     CNN                       - Create a CNN (conv, pool, conv, pool, FC, output)
                                 configuration for MNIST data processing.
     ConvLayer                 - Create a convolutional layer with specified weights,
                                 bias and activation function
     Conv2D                    - Create a 2D convolutional layer for a CNN
     MaxPool                   - Create a maxpool layer for a CNN specifying the
                                 size=strides
     MaxPool2x2                - Create a maxpool layer for a CNN with
                                 size=strides = [1,2,2,1]
     MaxPool3x3                - Create a maxpool layer for a CNN with
                                 size=strides = [1,3,3,1]

---------------------------------------------------------------------------------------
 * Multilayer perceptrons
---------------------------------------------------------------------------------------
     OutputNode                - Output node w^T.x + b
     MLPLayer                  - MLP layer
     MLP2                      - MLP with 2 hidden layers (uses DNN)
     MLP5                      - MLP with 5 hidden layers (uses DNN)
     DNN                       - Deep Network function to construct an arbitrary MLP
     TrainDeepNetwork          - Build and Train a deep network model

---------------------------------------------------------------------------------------
 * Recurrant Neural Networks
---------------------------------------------------------------------------------------
     RNN                       - Create a Long Short Term Memory (LSTM) RNN [to do]

---------------------------------------------------------------------------------------
 * Training helper functions
---------------------------------------------------------------------------------------
     ImageLabelInputQueue      - Creates a tf FIFO queue object
                                 full of image label pairs
     GetCost                   - Interface function to cost variable
                                 weights
     GetLoss                   - Interface function to loss variable
     GetOptimiser              - Interface function to optimiser

---------------------------------------------------------------------------------------
 * Plotting functions
---------------------------------------------------------------------------------------
     PlotDataImage             - Draw an image
     PlotImage                 - TC: Plot an image (catches imshow exceptions)
     PlotMNISTImage            - Draw an MNIST image
     PlotMNISTExample          - Display a 2D image with known example and label values
     PrintDataExample          - Print out a 2D image
     PrintMNISTExample         - Print out a 2D image with known example and label values
     PlotAllFeatures           - Make plots of all features in the feature space
     PlotAllScatterPlots       - Make all unique combinations of scatter plot for a
                                 feature space
     PlotFeature               - Plot the specified feature in the data
     ScatterPlot               - Plot the specified features in the data
     SaveArrayImage            - TC: Saves a np.array to disk in an image format e.g. png

---------------------------------------------------------------------------------------
 * Calculations
---------------------------------------------------------------------------------------
     CalcCorrelationMatrix     - Compute the (Pearson) correlation matrix for a data
                                 sample
     CalcCovarianceMatrix      - Compute the covariance matrix for a data sample
     CalcCovariance            - Compute the covariance between two columns of a data
                                 sample
     CalcCorrelationMatrixT    - Compute the (Pearson) correlation matrix for a tensor
     CalcCovarianceMatrixT     - Compute the covariance matrix for a tensor
     CalcCovarianceT           - Compute the covariance between two columns of a tensor
     CalcMeansAndRMS           - Compute the mean and RMS values of the features in the
                                 data
     PrintMatrix               - Print a correlation or covariance matrix

---------------------------------------------------------------------------------------
 * Data I/O
---------------------------------------------------------------------------------------
     ArraySplit                - Splits an array into partitions.
     CutOnData                 - Apply a numerical cut on the data
     CutOutNegativeData        - Remove examples with negative values from the data
                                 (e.g. use this function to remove examples with
                                 negative weights prior to building a feature set).
     ExtractColumn             - Extract a sigle column from a tf tensor of data
     ExtractColumns            - Extract a subspace of the input feature space from a
                                 tf tensor of data, specifying subspace order
     ExtractColumnTensor       - Extract a sigle column from a tf tensor of data
     ExtractColumnsTensor      - Extract a subspace of the input feature space from a
                                 tf tensor of data, specifying subspace order
     GetMNISTData              - Function to read in MNIST data examples
     MakeDir                   - Make a directory for output
     MergeFeatureSets          - Combine data samples to make a set that presents
                                 different labels in order (s/b/s/b/ etc).
     parseOptionsIO            - Option passing function; decode if VERBOSE or QUITE
                                 print out is required for I/O functions.
     PrintDataTFV              - Print the data to screen for inspection
     ReadCSV                   - Read in a CSV data file
     ReadCSVasTFV              - Read in a CSV data file and convert the output to a tf
                                 variable with shape [Ndata, dim(x)]
     WriteCSV                  - Write data to a CSV file
     WriteTFVasCSV             - Convert the [Ndata, dim(x)] tf variable to a python
                                 array and then output to a CSV data file
     FilenameList              - TC: Walk through a source directory and generate a list of
                                 files
     ImageLabelMap             - TC: Generates a .txt map from image paths to their labels
     ReadImageLabelMap         - TC: Reads a .txt image label map
     ReadImagesFromDisk        - TC: Reads images from disk
     WriteTFRecord             - TC: Writes a .tfrecords file of images and labels
                                 includes unlisted helper functions.
                                 NB: Poorly named does not work for general data
     TFRecordRead              - TC: Reads a .tfrecords file of images and labels
                                 includes unlisted helper functions.
                                 NB: Poorly named does not work for general data
     ReadImage                 - Reads image files from disk using PIL
     WriteNPY                  - Writes numpy arrays to disk (.npy files)
     ReadNPY                   - Reads numpy arrays from disk (.npy files)

---------------------------------------------------------------------------------------
 * Data pre-processing
---------------------------------------------------------------------------------------
     GaussianiseColumn         - Map the data into a Gaussian-like feature space
     GetFeatureMaxMin          - Determine the range of features for a specified column
     GetFeatureRange           - Determine the scale and offset to normalise features
                                 for a specified column
     NormaliseColumn           - Normalise a column to the specified range:
                                 default [0, 1]
     NormaliseData             - Normalise data columns to the specified range:
                                 default [0, 1]
     NormaliseDataNeg1To1      - Normalise data columns to the range [-1.0, 1.0]
     CrudeDownSample           - Downsample an image(array) by a factor of 1/n (n >= 2)
     SplitLabelledData         - Split some labelled data into train/test.
     OneHotConversion          - Converts labels to one-hot vectors.
     BinaryClasses             - Converts multi-class labels to binary class labels.

---------------------------------------------------------------------------------------
 * Drawing Functions
---------------------------------------------------------------------------------------
     ImGen                     - Generates a null image of specified shape
     InCircle                  - Checks if a coordinate (x, y) is in a specified circle
     InSquare                  - Checks if a coordinate (x, y) is in a specified square
     InRing                    - Checks if a coordinate (x, y) is in a specified ring
     DrawCircles               - Draw circles on a given image specified by several
                                 lists of parameters
     NoiseGen                  - Generates uniform random noise in a specified range of
                                 intensities on a given image

---------------------------------------------------------------------------------------
 * Miscellaneous functions
---------------------------------------------------------------------------------------
     TestResult                - Print the outcome of a boolean flag test
                                 (for tester scripts)
     TensorBoardInit           - Initiliase the tensorboard log writer
     TensorBoardFinalise       - Finalise graph logging

     CurrentTime               - Gets the current time.
     TimeInterval              - Computes the interval between now and a specified time.
     PickElement               - Gets a random element from a list
---------------------------------------------------------------------------------------
 Notes:
---------------------------------------------------------------------------------------
 * This hipster is written for a single CPU/GPU set of computations.  This is not optimal
   for systems where multiple GPUs are available.

 * Weight parameter initialisation:
   All nodes, layers and networks constructed using this
   hipster should use the WeightVariable and BiasVariable initialisers to ensure
   that these hyperparameters are set to non trivial values before the optimisation
   commences.  This is especially important for complicated network configurations.
   If you construct your own network the same principles apply and the same functions
   should be used.  The random initialisers are tuned to facilitate the use of
   relu activation functions.

 * Dropout:
   When using drop out to promote hyperparameter generalisation note that
   an increase in training epochs may be required in order to obtain equivalent
   accuracy for the network c.f. with a keep probability of 1.0.  If the
   keep probability is set to 1.0 then dropout nodes are not included in the
   graph on construction; so if you intend to build a model and iteratively
   change the dropout fraction, then use an initial keep probability less
   than unity.

---------------------------------------------------------------------------------------
 Adrian Bevan (a.j.bevan@qmul.ac.uk), Copyright (C) QMUL 2017
=======================================================================================
"""

print("Welcome to HIPSTER - Heavily Ionising Particle Standard Toolkit for Event Recognition")

# import all dependencies at this stage
import os
#os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
from PIL import Image
import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
import string
import math
import gzip
import uuid
import random
import time
from tensorflow.contrib.learn.python.learn.datasets.mnist import read_data_sets

"""
=======================================================================================
 * Class Definitions
=======================================================================================
"""

class colours:
    """
    -------------------------------------------------------------------------------------
     Class that defines colours for print statements, including info warning and error.
    -------------------------------------------------------------------------------------
    """
    PURPLE = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    INFO = GREEN + BOLD + "INFO:" + ENDC + " " #These are also examples of use cases
    WARNING = YELLOW + BOLD + "WARNING:" + ENDC + " "
    ERROR = RED + BOLD + "ERROR:" + ENDC + " "

class MLPTrainingOptions:
    """
    -------------------------------------------------------------------------------------
     Training options class, to facilitate passing setting up and passing
     training options to training methods.

     This contains the options required for a MLP
    -------------------------------------------------------------------------------------
    """
    def __init__(self, name):
        self.name               = name
        self.learning_rate      = 0.01
        self.training_epochs    = 25
        self.display_step       = 5
        self.batch_size         = 100
        self.n_outputs          = 1
        self.NTrainingData      = 5000
        self.ActivationFunction = "softmax"
        self.DropOutKeepProb    = 1.0
        self.UseWeights         = bool(0)
        self.TrainingPlotDir    = "DNNtrain/"
        self.Verbose            = bool(0)

    def set_learning_rate(self, rate):
        self.learning_rate = rate

    def set_epochs(self, epoch):
        self.training_epochs = epoch

    def set_display_step(self, step):
        self.display_step = step

    def set_batch_size(self, size):
        self.batch_size = size

    def set_NTrainingData(self, N):
        self.NTrainingData = N

    def set_activation_function(self, fn):
        self.ActivationFunction = fn

    def set_DropOutKeepProb(self, frac):
        self.DropOutKeepProb = frac

    def set_UseWeights(self, useWeights):
        self.UseWeights = useWeights

    def Print(self):
        print ("\tTraining options name = ", self.name)
        print ("\tLearning rate         = ", self.learning_rate)
        print ("\tN(epochs)             = ", self.training_epochs)
        print ("\tDisplay steps eac     = ", self.display_step, " epochs")
        print ("\tBatch size            = ", self.batch_size, " examples")
        print ("\tN(training data)      = ", self.NTrainingData)
        print ("\tActivationFunction    = ", self.ActivationFunction)
        print ("\tDrop Out Keep Prob    = ", self.DropOutKeepProb,
               " (= 1 - Drop Out Rate)")
        print ("\tUse example weights   = ", self.UseWeights)


class CNNTrainingOptions:
    """
    -------------------------------------------------------------------------------------
     Training options class, to facilitate passing setting up and passing
     training options to training methods.

     This contains the options required for a CNN
    -------------------------------------------------------------------------------------
    """
    def __init__(self, name):
        self.name               = name
        self.type               = "standard" # n*(conv-pool) + fully connected layer
                                             # configuration is the default
        self.learning_rate      = 1e-4
        self.training_epochs    = 25
        self.display_step       = 5
        self.batch_size         = 100
        self.n_outputs          = 1
        self.NTrainingData      = 5000
        self.ActivationFunction = "relu"
        self.DropOutKeepProb    = 1.0
        self.TrainingPlotDir    = "CNNtrain/"
        self.Verbose            = bool(0)
        self.image_x            = 28   # default initialiser works for MNIST
        self.image_y            = 28   # default initialiser works for MNIST
        self.convLayers         = []
        self.poolLayers         = []
        self.imageShape         = []
        self.fullyConnectedLayer= []


    def autoconfigure(self):
        """
        ---------------------------------------------------------------------------------
         Autoconfigure the CNN.  This is implemented based on the type of CNN
          architecture.
        ---------------------------------------------------------------------------------
        """
        ScaleX = float(self.image_x)
        ScaleY = float(self.image_y)
        Build = bool(1)
        InputFilterDepth = self.imageShape[3]
        OutputFilterDepth = 64
        while(Build):
            if self.type =="standard":
                # Prepare conv and pool layers based on 5x5 filters for the conv
                # layers and
                thisConv = [5,5,InputFilterDepth, OutputFilterDepth]
                thisPool = [1,2,2,1]

                ScaleX = ScaleX/thisPool[1]
                ScaleY = ScaleY/thisPool[2]
                if (not ( int(ScaleX) == ScaleX)) or (not ( int(ScaleY) == ScaleY)):
                    Build = 0
                else:
                    self.convLayers.append(thisConv)
                    self.poolLayers.append(thisPool)
            else:
                print ("Unknown CNN type - CNNTrainingOptions unable to autoconfigure")

        # Prepare the fully connected layer configuation
        lastLayer = self.convLayers[-1]
        self.fullyConnectedLayer = [int(ScaleX)*int(ScaleY)*lastLayer[-1], 1024]



    def reconfigure(self):
        """
        ---------------------------------------------------------------------------------
         Determine how many layers of conv-pool work with this image size
         and maxpool filter configuration.  Remove surplus layers that
         are not possbile to accomodate.
        ---------------------------------------------------------------------------------
        """
        Npool = len(self.poolLayers)
        ScaleX = float(self.image_x)
        ScaleY = float(self.image_y)
        for i in range(Npool):
            thisPool = self.poolLayers[i]
            ScaleX = ScaleX/thisPool[1]
            ScaleY = ScaleY/thisPool[2]

            # Review configuration and if the configuration is not viable, remove the
            # offending layers and reconfugure the FC layer.
            if (not ( int(ScaleX) == ScaleX)) or (not ( int(ScaleY) == ScaleY)):
                print ("""
*******************************************************************************
The CNN configuration is not viable - there are too many conv-pool layers requested
for this model.  The model will be reconfigured to remove the untennable layers,
and the input nodes to the fully connected layer will be recomputed to accomodate
this modification.

The following conv/pool configuration was originally requested:
""")
                print (len(self.convLayers), " convolution layers with shapes:")
                print (self.convLayers)
                print (len(self.poolLayers), " convolution layers with shapes:")
                print (self.poolLayers)

                print ("This CNN will be reconfigured to have ", Npool - i - 1,
                       " conv-pool units")
                print ("""
*******************************************************************************
""")
                for j in range(Npool - i):
                    self.convLayers.pop()
                    self.poolLayers.pop()

                self.fullyConnectedLayer=[ self.get_FCinputNodes(),
                                          self.fullyConnectedLayer[1] ]
                break


    def check_configuration(self):
        """
        ---------------------------------------------------------------------------------
         check that the image size is reduced in a consistent way by the maxpool
         layers (i.e. that it is possible to resize correctly).
        ---------------------------------------------------------------------------------
        """
        Npool = len(self.poolLayers)
        ScaleX = float(self.image_x)
        ScaleY = float(self.image_y)
        for i in range(Npool):
            thisPool = self.poolLayers[i]
            ScaleX = ScaleX/thisPool[1]
            ScaleY = ScaleY/thisPool[2]

        # Review configuration and print out warnings for issues found
        if (not ( int(ScaleX) == ScaleX)) or (not ( int(ScaleY) == ScaleY)):
            error_message = """
*******************************************************************************
CNN CONFIGURATION ERROR: The maxpool reduction of the image is attempting to
create a non-integer number of pixels.  Please review the configuration as
this will not work.
*******************************************************************************
"""
            print (colours.ERROR + error_message)
            print ("\tN(pixels in x) = ", ScaleX)
            print ("\tN(pixels in y) = ", ScaleY)
            print ("\tImage size x*y = ", self.image_x, " * ", self.image_y)
            print ("\tPool layers    = ", self.poolLayers)
            return 1
        else:
            print (colours.INFO + "CNN configuration is correctly setup.")
        return 0



    def get_FCinputNodes(self):
        """
        ---------------------------------------------------------------------------------
        assumes that pool="SAME" is used for the conv layers, so that only the pool
        layers reduce the image size.  The FCinput layer has to connect to each voxel
        in the image volume produced by the last layer. This is has the volume depth of
        the last conv layer, and the lateral dimension is the original image size
        reduced by the max pool layer reduction factors.
        (depends on the pool layer sizes).
        ---------------------------------------------------------------------------------
        """
        lastConvLayer = self.convLayers[-1]
        FCinputNodes = lastConvLayer[-1]*self.image_x*self.image_y
        for i in range(len(self.poolLayers)):
            maxPoolLayer = self.poolLayers[i]
            FCinputNodes = FCinputNodes/maxPoolLayer[1]/maxPoolLayer[2]
            if not ( int(FCinputNodes) == FCinputNodes):
                print ("""
*******************************************************************************
CNN CONFIGURATION ERROR - The number of FC input nodes expected is not integer
                          This config error must be fixed before proceeding.
""")
        return int(FCinputNodes)


    def Print(self):
        """
        ---------------------------------------------------------------------------------
         Print function for the CNN options
        ---------------------------------------------------------------------------------
        """
        print ("Instance of CNNTrainingOptions called ", self.name)
        print ("\tLearning rate           = ", self.learning_rate)
        print ("\tEpochs                  = ", self.training_epochs)
        print ("\tDisplay step            = ", self.display_step       )
        print ("\tBatch size              = ", self.batch_size         )
        print ("\tNumber of outputs       = ", self.n_outputs          )
        print ("\tNumber of training data = ", self.NTrainingData      )
        print ("\tActivation function     = ", self.ActivationFunction )
        print ("\tDrop out keep prob      = ", self.DropOutKeepProb,
               " (drop out prob = ", 1-self.DropOutKeepProb, ")")
        print ("\tTraining Plot Directory = ", self.TrainingPlotDir    )
        print ("\tVerbose                 = ", self.Verbose            )
        print ("\tImage x size            = ", self.image_x            )
        print ("\tImage y size            = ", self.image_y            )
        print ("\tConv Layers             = ", self.convLayers         )
        print ("\tMaxpool layers          = ", self.poolLayers         )
        print ("\tImage Shape             = ", self.imageShape         )
        print ("\tFully Connected Layer   = ", self.fullyConnectedLayer)






class Fisher:
    """
    -------------------------------------------------------------------------------------
     Class to facilitate computation of fisher coefficients for an MVA problem.

     alpha = W^-1 (mu[A] - mu[B])

    -------------------------------------------------------------------------------------
    """
    def __init__(self, name="fisher"):
        self.name               = name
        self.alpha              = []
        self.sess               = tf.Session()
        self.deltaMu            = []
        self.W                  = []
        self.Winv               = []
        self.Verbose            = bool(0)

    def CalcCoefficients(self, classA, classB, columns=[]):
        """
        ---------------------------------------------------------------------------------

         Function to compute the fisher discriminant coefficients given tensors of the
         specified shape corresponding to two classes of event
         [N examples, feature sapace dimension] first reset the arrays to make sure that
         we don't overwrite anything from a previous computation; then proceed with
         computation.
        ---------------------------------------------------------------------------------
        """
        del self.alpha[:]
        del self.deltaMu[:]
        del self.W[:]
        del self.Winv[:]

        init = tf.global_variables_initializer()
        self.sess.run(init)
        meanA, varA = tf.nn.moments(classA, axes=[0])
        meanB, varB = tf.nn.moments(classB, axes=[0])

        if(len(columns) == 0):
            # make a dummy set of columns for covariance matrix comp
            row0 = self.sess.run(classA)[0]
            for i in range(len(row0)):
                columns.append(str(i))

        # compute the covariance matrices for the data
        CovA = CalcCovarianceMatrixT(self.sess, classA, columns)
        CovB = CalcCovarianceMatrixT(self.sess, classB, columns)

        Winv = tf.matrix_inverse(tf.add(CovA, CovB))
        DeltaMu = tf.subtract(meanA, meanB)

        # compute coefficients using array form
        self.Winv     = self.sess.run(Winv)
        self.deltaMu  = self.sess.run(DeltaMu)

        for i in range(len(self.Winv)):
            row = self.Winv[i]
            thisAlpha = 0
            for j in range(len(row)):
                thisAlpha += row[j]*self.deltaMu[j]
                self.alpha.append(thisAlpha)

        if(self.Verbose):
            if(len(columns)>0):
                for i in range(len(columns)):
                    print ("\talpha({0}) = {1:.4f}".format(columns[i], self.alpha[i]))
            else:
                for i in range(len(columns)):
                    print ("\talpha({0:d}) = {1:.4f}".format(i, self.alpha[i]))


    def CalcVal(self, dataExample):
        """
        ---------------------------------------------------------------------------------
         Given the computed coefficients, evaluate the fisher output for an
         example provided
        ---------------------------------------------------------------------------------
        """
        thisVal = 0
        if(len(self.alpha) >0):
            for j in range(len(dataExample)):
                thisVal += dataExample[j]*self.alpha[j]
        else:
            print ("""
ERROR - unable to compute Fisher discriminant as coeffients have not yet been computed"

To use the function Fisher::CalcVal, you first have to call CalcCoefficients to evaluate
the Fisher coefficients.
""")
        return thisVal


    def Print(self):
        """
        ---------------------------------------------------------------------------------
         Print out the Fisher coefficients, deltaMu and W^-1
        ---------------------------------------------------------------------------------
        """
        print ("Fisher computation summary:")
        if(len(self.alpha) == 0):
            print ("""
ERROR - unable to display Fisher discriminant as coeffients have not yet been computed"

To use the function Fisher::Print, you first have to call CalcCoefficients to evaluate
the Fisher coefficients.
""")
            return

        print ("Delta Mu column tensor:")
        for i in range(len(self.deltaMu)):
            print ("\tDeltaMu({0:d}) = {1:.4f}".format(i, self.deltaMu[i]))

        print ("W^-1 tensor:")
        PrintMatrix(self.Winv)

        print ("\nalpha coefficients:")
        for i in range(len(self.alpha)):
            print ("\talpha({0:d}) = {1:.4f}".format(i, self.alpha[i]))


class ROC:
    """
    -------------------------------------------------------------------------------------
     TC
      Class to calculate ROC curves and display in TensorBoard

     Inputs:
              predict_label_file             - .txt files of predictions and labels
              threshold_num                  - the number of thresholds to plot
              save_dir                       - where the summary writer saves to

     Output can be viewed with TensorBoard in localhost:6006

     KF
     added functions:
     calc_auc() - !should only be called after calc()!
                  calculates area under curve
       arguments:
         self: uses self.fpr and self.tpr
       returns:
         self.auc: returns area under curved from updated self.auc

     plot_roc() - !should only be called after calc()!
                  makes matplotlib plot of ROC curve
       arguments:
         self: uses self.fpr, self.tpr, self.auc
         save=False: keyword argument, boolean, False only shows image, True saves it
         plot_dir=None: keyword argument, takes output directory for the plot,
                        only used if save=True
    -------------------------------------------------------------------------------------
    """
    def __init__(self, predict_label_file, threshold_num, save_dir):
        '''''file format: dataid,predict_score,label
        the predict_score should be between 0 and 1
        the label should be 0 or 1
        threshold_num: number of threshold will plot'''
        #load predict_label_file to predicts and labels
        print(colours.INFO + "reading predict label file", predict_label_file)
        fd = open(predict_label_file)
        fdl = fd.readline()
        self.predicts = []
        self.labels = []
        self.total = 0
        while len(fdl) > 0:
            val = fdl.split(',')
            self.predicts.append(float(val[1]))
            self.labels.append(True if int(val[2]) == 1 else False)
            fdl = fd.readline()
            self.total += 1
        fd.close()
        self.threshold_num = threshold_num
        self.trues = 0 #total of True labels
        self.fpr = [] #false positive
        self.tpr = [] #true positive
        self.ths = [] #thresholds
        self.save_dir = save_dir #directory of the TensorBoard plot
        self.auc = 0 #area under curve
        self.writer = tf.summary.FileWriter(self.save_dir)

    def calc(self):
        for label in self.labels:
            if label:
                self.trues += 1
        threshold_step = 1. / self.threshold_num
        for t in range(self.threshold_num + 1):
            th = 1 - threshold_step * t
            tn, tp, fp, fpr, tpr = self._calc_once(th)
            self.fpr.append(fpr)
            self.tpr.append(tpr)
            self.ths.append(th)
            self._save(fpr, tpr)
        #print(self.fpr)
        #print(self.tpr)
        #print(self.ths) # Use these to plot elsewhere also

    def _save(self, fpr, tpr):
        summt = tf.Summary()
        summt.value.add(tag="roc", simple_value = tpr)
        self.writer.add_summary(summt, fpr * 100) #for tensorboard step drawable
        self.writer.flush()

    def _calc_once(self, t):
        fp = 0
        tp = 0
        tn = 0
        for i in range(self.total):
            if not self.labels[i]:
                if self.predicts[i] >= t:
                    fp += 1
                else:
                    tn += 1
            elif self.predicts[i] >= t:
                tp += 1
        fpr = fp / float(fp + tn) #precision
        #fpr = fp / float(fp + tp) #detection
        tpr = tp / float(self.trues)
        return tn, tp, fp, fpr, tpr

    def plot_roc(self, save=False, plot_dir=None):
        if save == True:
            if plot_dir is not None:
                self.calc_auc()
                plt.figure()
                plt.title('Receiver Operating Characteristic')
                plt.xlabel('False positive rate')
                plt.ylabel('True positive rate')
                plt.xlim([0,1])
                plt.ylim([0,1])
                plt.plot([0]+self.fpr, [0]+self.tpr, 'b', label = "Area Under Curve = %f" % self.auc)
                plt.plot([0,1],[0,1],'r--')
                plt.legend(loc = 'lower right')
                plt.savefig(plot_dir + '/roc.pdf')
            else:
                raise ValueError('No directory set for saving the ROC curve plot. \n\t\t\tPlease use "plot_dir=some/directory" when calling the function.')
        else:
            self.calc_auc()
            plt.figure()
            plt.title('Receiver Operating Characteristic')
            plt.xlabel('False positive rate')
            plt.ylabel('True positive rate')
            plt.xlim([0,1])
            plt.ylim([0,1])
            plt.plot([0]+self.fpr, [0]+self.tpr, 'b', label = "Area Under Curve = %f" % self.auc)
            plt.plot([0,1],[0,1],'r--')
            plt.legend(loc = 'lower right')
            plt.show()

    def calc_auc(self):
        self.auc = np.trapz(self.tpr, self.fpr)
        return self.auc


"""
========================================================================
 * General NN operations
 General helper functions follow on from here; these are algorithm
 non-specific functions.

 When constructing large networks the weights and biases should be initialised
 with a non-trivial set of values. This is achieved here using a truncated
 normal distribution which is motiviated by pathologies that exist for the
 relu activiation function used for CNNs. Having non-zero hyperparameters
 will give the optimisation algorithm a unique starting point to converge
 on a solution.
========================================================================
"""
def CountLabels(labels):
    """
    ----------------------------------------------------------------------
    Determine how many unique labels are in the labels tensor. Use this
    to determine the number of output classes to avoid hard coding
    that number into the hipster tools and derived examples.

    Having done that, create a tensor that can be used in training for
    this set of labels.
    ----------------------------------------------------------------------
    """
    Unique = []

    print ("Counting the number of unique labels")
    # loop over the labels and note those that are unique
    thisData = labels
    for i in range(len(thisData)):
        thisLabel = thisData[i]
        if i == 0:
            Unique.append( thisLabel )
        else:
            KnownLabel = 0
            for j in range(len(Unique)):
                if thisLabel == Unique[j] :
                    KnownLabel = 1
            if KnownLabel == 0:
                Unique.append( thisLabel )

    print ("There are ", len(Unique),
           " unique labels in this data sample; expanding the label tensor dimensionality")
    theNewLabels = []
    for i in range(len(thisData)):
        thisLabel = []
        for j in range(len(Unique)):
            if j == thisData[i]:
                thisLabel.append( 1 )
            else:
                thisLabel.append( 0 )

        theNewLabels.append( thisLabel )

        theNewLabelsTensor = tf.Variable(theNewLabels, name="labels_tensor")

    return len(Unique), theNewLabelsTensor



def PrepareWeights( Weights, Labels, n_classes ):
    """
    ----------------------------------------------------------------------
    prepare a tensor of weight values, where the event weight corresponds to
    the entry in the corresponding label column so that the shape of the
    output tensor is given by [?, n_classes], in analogy with the
    CountLabels function above.

    Weights is an array of weights (rank is 1)
    Labels is an array of weights (rank is 1)
    n_classes is the number of distinct classes
    ----------------------------------------------------------------------
    """
    sumOfWeights = 0
    theNewWeights = []
    for i in range(len(Weights)):
        thisWeight = []
        for j in range(n_classes):
            if j == Labels[i]:
                thisWeight.append( Weights[i] )
                sumOfWeights += Weights[i]
            else:
                thisWeight.append( 0 )

        theNewWeights.append( thisWeight)

        theNewWeightsTensor = tf.Variable(theNewWeights, name="weights_tensor")

    return theNewWeightsTensor, sumOfWeights




def WeightVariable(shape, thisname="weight"):
    """
    ----------------------------------------------------------------------
     function to create weights with a randomly chosen initialisation.
    ----------------------------------------------------------------------
    """
    if not (thisname=='weight'):
        thisname += "_weight"

    initial = tf.truncated_normal(shape, stddev=0.1)
    return tf.Variable(initial, name=thisname)

def BiasVariable(shape, thisname="bias"):
    """
    ----------------------------------------------------------------------
     function to create biases with a randomly chosen initialisation.
    ----------------------------------------------------------------------
    """
    if not (thisname=="bias"):
        thisname += "_bias"

    initial = tf.constant(0.1, shape=shape)
    return tf.Variable(initial, name=thisname)


def NetworkActivationFunction(argument, activation_function):
    """
    ----------------------------------------------------------------------
     Allow for different types of activation function to be used when building
     a layer.  Possible options are:
       sigmoid
       tanh
       elu
       softplus
       softsign
       softmax
       relu
       relu6
       relu_x

     If the function name is incorrectly specified a warning is given and the default value
     of a sigmoid activation function will be adopted.

     KF
     Scope was added for graphing and making histograms in TensorBoard.
     This does not affect projects not using TensorBoard.
    ----------------------------------------------------------------------
    """
    if(activation_function == "relu"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.relu(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "relu6"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.relu6(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "relu_x"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.relu_x(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "sigmoid"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.sigmoid(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "tanh"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.tanh(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "elu"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.elu(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "softplus"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.softplus(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "softsign"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.softsign(argument)
            tf.summary.histogram('activations', layer)
    elif(activation_function == "softmax"):
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.softmax(argument)
            tf.summary.histogram('activations', layer)
    else:
        # mis-configured layer; make sure that the error is flagged to the user
        print ("Unknown activation function type ", activation_function, " using a sigmoid instead!")
        with tf.name_scope('Activation') as scope:
            layer       = tf.nn.sigmoid(argument)
            tf.summary.histogram('activations', layer)
    return layer


"""
========================================================================
 Helper functions for convolutional neural networks follow on from
 here.  CNNs require convolution and pool layers.  There are generic
 functions that facilitate the creation of these layers, along with
 those hardwired to generate specific configurations.
========================================================================
"""
def BuildCNN(x, opt):
    """
    ----------------------------------------------------------------------
     Build a CNN model depending on the CNNTrainingOptions
    ----------------------------------------------------------------------
    """
    model = 0

    # Always reconfigure the CNNTrainingOptions to a viable model form before trying to construct the
    # model. This removes the need for the user to make this check.
    print(colours.INFO + 'Building CNN')
    opt.reconfigure()

    if opt.type == "standard":
        model = CNN(x, opt.DropOutKeepProb, opt.convLayers, opt.poolLayers, opt.fullyConnectedLayer, opt.imageShape, opt.n_outputs, opt.name, opt.ActivationFunction)
    else:
        error_message = """
BuildCNN ERROR - unable to construct the CNN model of specified type.  See the configuration below and rectify.
The known CNN types are:
   standard
"""
        print (colours.ERROR + error_message)
        opt.Print()

    return model


def Conv2D(x, W):
    """
    ----------------------------------------------------------------------
     Create a 2D convolutional layer.  By default padding is set to "same"
     which preserves the image size.
    ----------------------------------------------------------------------
    """
    return tf.nn.conv2d(x, W, strides=[1, 1, 1, 1], padding='SAME')


def MaxPool(x, size=[1, 2, 2, 1]):
    """
    ----------------------------------------------------------------------
     create a max pooling kernel: n x n with strides matching the max pool layer size
     to avoid overlapping the pooling stage.
    ----------------------------------------------------------------------
    """
    return tf.nn.max_pool(x, ksize=size, strides=size, padding='SAME')


def MaxPool2x2(x):
    """
    ----------------------------------------------------------------------
     create a max pooling kernel: 2 x 2 with strides to avoid overlapping the
     pooling stage.
    ----------------------------------------------------------------------
    """
    return max_pool(x, [1, 2, 2, 1])


def MaxPool3x3(x):
    """
    ----------------------------------------------------------------------
    create a max pooling kernel: 3 x 3 with strides to avoid overlapping the
    pooling stage.
    ----------------------------------------------------------------------
    """
    return max_pool(x, [1, 3, 3, 1])


def ConvLayer(x, W, b, activation_function="relu"):
    """
    ----------------------------------------------------------------------

     Construct a convolutional layer for some image x, with weights W, bias b and the
     specified activation function.  The default is to use the relu activiation function.

     KF
     Scope was added for graphing and making histograms in TensorBoard.
     This does not affect projects not using TensorBoard.
    ----------------------------------------------------------------------
    """
    with tf.name_scope('Wx_plus_b') as scope:
        argument = Conv2D(x, W) + b
        layer = NetworkActivationFunction(argument, activation_function)
        tf.summary.histogram('pre_activations', argument)
    return layer


def CNN(x, keep_prob, convLayers, poolLayers, fullyConnectedLayer, imageShape, n_outputs, name="cnn", activation_function="relu"):
    """
    ----------------------------------------------------------------------
    Build a convolutional network with two lots of "convolutional layers followed by
    pool layers".  Padding is set to the same and the keep probability is tunable.
    The output of the second pool layer is fed into a fully connected layer that
    in turn is passed to the output node.

     The default activation function is a relu activation function (also configurable).
     To call this function the image data needs to be provided along with a
     tf.placeholder (dropout keep probability).  The name and activation functions
     are optional.

     Example: MNIST data with 2 sets of conv+pool layers, a flat fully connected layer and 10 output types

       keep_prob = tf.placeholder(tf.float32)

       convLayers = []
       convLayers.append([5, 5, 1, 32])
       convLayers.append([5, 5, 32, 64])

       poolLayers = []
       poolLayers.append([1, 2, 2, 1])
       poolLayers.append([1, 2, 2, 1])

       fullyConnectedLayer = [7 * 7 * 64, 1024]
       imageShape = [-1, 28, 28, 1]

       y_conv = hipster.CNN(x, keep_prob, convLayers, poolLayers, fullyConnectedLayer, imageShape, 10)

     The CNN built will be configured according to the inputs; there will be some number of
     convolutional layers followed by maxpool layers; then there is a final fully connected
     layer (with dropout) before the output node is constructed.

     KF
     TensorBoard support was added for visualisations of convolutional layer filters, histograms and graphs.
     These changes do not affect projects which are not using TensorBoard.
    ----------------------------------------------------------------------
    """
    if(len(convLayers) != len(poolLayers)):
        if( len(poolLayers) > 0):
            print ("\tMismatch in the number of convolution and pool layers for this CNN, can't build a model")
            return 0
        else:
            print ("\tThe number of pool layers is set to zero for this CNN - no maxpooling")

    print (colours.INFO + "Will create a convolutional neural network with:")
    print ("\t", len(convLayers), " convolutional layers")
    print ("\t", len(poolLayers), " maxpool layers")

    name_base = name + "layer"
    CNNConvLayers     = []
    CNNPoolLayers     = []
    CNNLayers_dropout = []
    layer_weights     = []
    layer_biases      = []

    # reshape the input image to the specified shape
    x_image = tf.reshape(x, imageShape)
    theImage = x_image

    # build the CNN, start with a convolution layer, follow this with a maxpool layer
    for layer in range(len(convLayers)):
        shape = convLayers[layer]
        print ("\tCreating convolutional layer number ", layer, "with shape ", shape)
        with tf.name_scope('conv' + str(layer+1)):
        # create the biases and weights for the convolutional layer
            with tf.name_scope('weights'):
                theseWeights = WeightVariable(shape)
                layer_weights.append(theseWeights)
                with tf.name_scope('summaries'):
                    mean = tf.reduce_mean(theseWeights)
                    tf.summary.scalar('mean', mean)
                    with tf.name_scope('stddev'):
                        stddev = tf.sqrt(tf.reduce_mean(tf.square(theseWeights - mean)))
                    tf.summary.scalar('stddev', stddev)
                    tf.summary.scalar('max', tf.reduce_max(theseWeights))
                    tf.summary.scalar('min', tf.reduce_min(theseWeights))
                    tf.summary.histogram('histogram', theseWeights)

            with tf.name_scope('biases'):
                theseBiases = BiasVariable([shape[3]])
                layer_biases.append(theseBiases)
                with tf.name_scope('summaries'):
                    mean = tf.reduce_mean(theseBiases)
                    tf.summary.scalar('mean', mean)
                    with tf.name_scope('stddev'):
                        stddev = tf.sqrt(tf.reduce_mean(tf.square(theseBiases - mean)))
                    tf.summary.scalar('stddev', stddev)
                    tf.summary.scalar('max', tf.reduce_max(theseBiases))
                    tf.summary.scalar('min', tf.reduce_min(theseBiases))
                    tf.summary.histogram('histogram', theseBiases)

            # convolution of reshaped image with the weight tensor and apply the
            # bias offset.
            thisConvLayer = ConvLayer(theImage, theseWeights, theseBiases, activation_function)
            theImage = thisConvLayer

            if(len(poolLayers) > 0):
                print ("\t         maxpool layer ", layer, " with shape ", poolLayers[layer])
                # Now the convolutional layer is maxpooled
                thisPoolLayer = MaxPool(thisConvLayer, poolLayers[layer])

                # update the image to be the output of the previous pool layer
                theImage = thisPoolLayer

        # Function allowing for visualisation in Tensorboard of filters applied
        # by the convolutional layers
        with tf.name_scope('Visualize_filters_' + str(layer+1)) as scope:
            # We concatenate the filters into one image of row size 8 images
            W1_a = theseWeights                       # i.e. [5, 5, 1, 32]
            W1_b = tf.split(W1_a, convLayers[layer][3], 3) # i.e. [32, 5, 5, 1, 1]
            rows = []
            for i in range(int((convLayers[layer][3])/8)):
                x1 = i*8
                x2 = (i+1)*8
                row = tf.concat(W1_b[x1:x2],0)
                rows.append(row)
            W1_c = tf.concat(rows, 1)
            c_shape = W1_c.get_shape().as_list()
            W1_d = tf.reshape(W1_c, [c_shape[2], c_shape[0], c_shape[1], 1])
            tf.summary.image("Visualize_kernels_" + str(layer+1), W1_d, 1024)
        # Save output from layer to TensorBoard
        with tf.name_scope('Image_output_conv' + str(layer+1)):
            image = thisConvLayer[0:1, :, :, 0:16]
            image = tf.transpose(image, perm=[3,1,2,0])
            tf.summary.image('Image_output_conv' + str(layer+1), image)



    print ("\tCreating fully connected layer with with shape ", fullyConnectedLayer)
    # Now we contruct a fully connected layer working from a 7x7 image with
    # 64 outputs and 1024 neurons.
    with tf.name_scope("Fully_Connected1") as scope:
        fullyConnectedLayerWeights = WeightVariable(fullyConnectedLayer)
        fullyConnectedLayerBiases  = BiasVariable([fullyConnectedLayer[1]])


        # fullyConnectedLayer take the image corresponding to the last pool layer and reshape to
        # the specified configuration [nodes given by the 1st element of fullyConnectedLayer
        fullyConnectedLayer_FlatImage = tf.reshape(theImage, [-1, fullyConnectedLayer[0]])
        fullyConnectedLayer_nodes = tf.nn.relu(tf.matmul(fullyConnectedLayer_FlatImage, fullyConnectedLayerWeights) + fullyConnectedLayerBiases)

        # use dropout as required
        fullyConnectedLayer_dropout = fullyConnectedLayer_nodes
        if(keep_prob != 1.0):
            fullyConnectedLayer_dropout = tf.nn.dropout(fullyConnectedLayer_nodes, keep_prob)

        print ("\tCreating readout layer with ", fullyConnectedLayer[1], " inputs, and ", n_outputs, " outputs")
        # this is the readout layer where we flatten the 1024 outputs of the previous
        # densely connected layer into evidence for the 10 possible types
        ReadoutLayerWeights = WeightVariable([fullyConnectedLayer[1], n_outputs])
        ReadoutLayerBiases = BiasVariable([n_outputs])
    with tf.name_scope("Final_matmul") as scope:
        y_conv = tf.matmul(fullyConnectedLayer_dropout, ReadoutLayerWeights) + ReadoutLayerBiases

    return y_conv


"""
========================================================================
 Helper functions for multilayer perceptrons and deep networks follow on
 from here.  These functions provide interfaces to TensorFlow's neural
 net functionality and aid creating an MLP of arbitrary configuraiton.
 The generic interface is the DNN function that allows one to create
 a deep network by specifying the input feature space dimensionality,
 the number of label types (or classes) and the hidden layer configuration.
 Two functions (MLP2 and MLP5) provide short hand interfaces to DNN.

 If keep_prob is not set to 1.0 then dropout will be used.
========================================================================
"""
def MLP2(x, n_input, n_hidden_1, n_hidden_2, n_classes, keep_prob=1.0, name="mlp", activation_function="sigmoid"):
    """
    ----------------------------------------------------------------------
     Construct 2 layer MLP that takes n_input features and outputs n_classes

     Network config: n_input:n_hidden_1:n_hidden_2:n_classes

       x:          feature value tensor
       n_input:    number of inputs (e.g. feature space dimensionality, or number of outputs from
                   a previous layer).
       n_hidden_1: number of nodes in layer 1.
       n_hidden_2: number of nodes in layer 2.
       n_classes:  number of output classes
       keep_prob:  The dropout keep probability (=1 to retain all nodes)
       name:       name for this layer (for identification in TensorBoard).
       activation_function: specify the activation function type (default value="sigmoid")

    ----------------------------------------------------------------------
    """
    layers = [n_hidden_1, n_hidden_2]
    output_layer = DNN(x, n_input, layers, n_classes, keep_prob, name, activation_function)

    return output_layer



def MLP5(x, n_input, n_hidden_1, n_hidden_2, n_hidden_3, n_hidden_4, n_hidden_5, n_classes, keep_prob=1.0, name="mlp", activation_function="sigmoid"):
    """
    ----------------------------------------------------------------------
     Construct 5 layer (deep) MLP that takes n_input features and outputs n_classes

     Network config: n_input:n_hidden_1:n_hidden_2:n_hidden_3:n_hidden_4:n_hidden_5:n_classes

       x:          feature value tensor
       n_input:    number of inputs (e.g. feature space dimensionality, or number of outputs from
                   a previous layer).
       n_hidden_1: number of nodes in layer 1.
       n_hidden_2: number of nodes in layer 2.
       n_hidden_3: number of nodes in layer 3.
       n_hidden_4: number of nodes in layer 4.
       n_hidden_5: number of nodes in layer 5.
       n_classes:  number of output classes
       keep_prob:  The dropout keep probability (=1 to retain all nodes)
       name:       name for this layer (for identification in TensorBoard).
       activation_function: specify the activation function type (default value="sigmoid")

    ----------------------------------------------------------------------
    """
    layers = [n_hidden_1, n_hidden_2, n_hidden_3, n_hidden_4, n_hidden_5]
    output_layer = DNN(x, n_input, layers, n_classes, keep_prob, name, activation_function)

    return output_layer


def DNN(x, n_input, layers, n_classes, keep_prob=1.0, name="DNN", activation_function="softmax"):
    """
    ----------------------------------------------------------------------
     Construct a deep MLP that takes n_input features and outputs n_classes.  The shape of the
     hidden layers is given by the list layers.

     Network config: n_input : hidden layers dictacted by the list layers : n_classes

       x:          feature value tensor
       n_input:    number of inputs (e.g. feature space dimensionality, or number of outputs from
                   a previous layer).
       layers:     list of number of inputs for the hidden layers
       n_classes:  number of output classes
       keep_prob:  The dropout keep probability (=1 to retain all nodes)
       name:       name for this layer (for identification in TensorBoard).
       activation_function: specify the activation function type (default value="sigmoid")

     Example: MNIST data sample with 784 dimensional feature space and 10 target classes,
     with a DNN with 5 hidden layers, each having 256 nodes.
       n_input = 784
       n_classes = 10
       layers = [256, 256, 256, 256, 256]
       output_layer = hipster.DNN(x, n_input, layers, n_classes, 1.0, "DNN", "softmax")

     If keep_prob is set to 1.0 then no drop out layers are constructed.
    ----------------------------------------------------------------------
    """
    name_base = name + "layer"
    n_prev = n_input
    x_prev = x
    DNNLayers = []
    DNNLayers_dropout = []

    print ("Deep neural network model; will create ", len(layers), " hidden layers")
    # construct the hidden layers
    for layer in range(len(layers)):
        thisName = name_base
        thisName += str(layer)
        thisN = layers[layer]
        thislayer = MLPLayer(x_prev, n_prev, thisN, thisName, activation_function)
        DNNLayers.append(thislayer)
        if keep_prob<1.0:
            thislayer_drop = tf.nn.dropout(thislayer, keep_prob)
            DNNLayers_dropout.append(thislayer_drop)
            x_prev = thislayer_drop
        else:
            x_prev = thislayer

        print ("\tCreated layer ", layer, " with ", n_prev, " inputs and ", thisN, " nodes, and name ", thisName)
        n_prev = thisN

    # construct the output layer from the final image of the transformed feature space
    if keep_prob<1.0:
        output_layer = OutputNode(DNNLayers_dropout[-1], layers[-1], n_classes)
    else:
        output_layer = OutputNode(DNNLayers[-1], layers[-1], n_classes)
    print ("\tCreated output layer with ", layers[-1], " inputs and ", n_classes, " outputs")

    return output_layer



def MLPLayer(x, n_input, n_hidden, name, activation_function="sigmoid"):
    """
    ----------------------------------------------------------------------
     Construct layer of perceptrons for a neural network; ensure that the weights are created
     with some random offset values.  The arguments input are:
       x:         feature value tensor
       n_input:   number of inputs (e.g. feature space dimensionality, or number of outputs from
                  a previous layer).
       n_hidden:  number of nodes in this layer.
       name:      name for this layer (for identification in TensorBoard).
       activation_function: specify the activation function type (default value="sigmoid")

    ----------------------------------------------------------------------
    """
    weight_name = name+"_weights"
    bias_name   = name+"_bias"
    w_layer     = tf.Variable(tf.random_normal([n_input, n_hidden]), name=weight_name)
    bias_layer  = tf.Variable(tf.random_normal([n_hidden]), name=bias_name)

    argument = tf.add(tf.matmul(x,w_layer),bias_layer)
    layer = NetworkActivationFunction(argument, activation_function)

    return layer



def OutputNode(x, n_input, n_output, thisname="outputnode"):
    """
    ----------------------------------------------------------------------
     Construct the output of the network, where the output layer combines the
     information from n_input nodes down into a space of evidences for the n_output
     possible classes in the problem.

     Here x is the value of the previous layer.
    ----------------------------------------------------------------------
    """
    thisWeight = thisname
    thisWeight += "_weight"
    thisBias = thisname
    thisBias += "_bias"
    output = tf.Variable(tf.random_normal([n_input, n_output]), name=thisWeight)
    bias_output = tf.Variable(tf.random_normal([n_output]), name=thisBias)
    output_layer = tf.matmul(x, output) + bias_output

    return output_layer




def TrainDeepNetwork(layers, FeatureSpace, Features, Labels, Weights, options):
    """
    ----------------------------------------------------------------------

     construct and train a deep network using the data provided

       layers        - the hidden layer structure for the DNN
       FeatureSpace  - tensor of the data
       Features      - list of column names for the feature space
       Labels        - labels for the training examples
       Weights       - event weights

     By default this uses an Adam optimiser, with softmax cross entropy
     The feature space dimensionality and number of classes for the output
     layer are determined from the data presented to the function.

    ----------------------------------------------------------------------
    """
    print ("""

    Will now build and train a deep network for training with this data sample

  """)

    # Network architecture parameters - the number of inputs is given by
    # the feature space dimensionality
    n_input = len(Features)

    # signal (H->tautau) and background (sum over all other backgrounds)
    # we need to expand the labels so that this is a tensor with size
    # [?, n_classes] in order to be feed into the model optimisation.
    # at the same time we can then use the dimensionality of a row in
    # the class label tensor to fix the number of output classes in the
    # model.  This allows us to have a generalised example to adapt to
    # other problems.
    # Initializing the variables
    sess = tf.Session()
    init = tf.global_variables_initializer()
    sess.run(init)

    n_classes, n_classesTensor        = CountLabels( sess.run(Labels) )
    n_weightsTensor, sumOfWeights     = PrepareWeights( sess.run(Weights), sess.run(Labels), n_classes )
    init = tf.global_variables_initializer()
    sess.run(init)

    # tf Graph input:
    #  x: is the tensor for the number of images being analysed (the placeholder
    #     entry None is used for that; and the number of features extracted from the image.
    #     In this case the features of the image correspond to a flattened representation of
    #     the pixels.  Which for this type of a network is not a problem.
    #
    #  y: is the tensor for the number of images being analysed and the corresponding
    #     labels for those images (corresponding to the numbers 0 through 9.
    #
    #  weight: is a placeholder for the event weight (used when analysing data that
    #     requires event weighting.
    #
    x         = tf.placeholder("float", [None, n_input],   name = "data_placeholder")
    y         = tf.placeholder("float", [None, n_classes], name = "label_placeholder")
    weight    = tf.placeholder("float", [None],            name = "weight_placeholder")
    keep_prob = tf.placeholder(tf.float32,                 name = "keep_prob_placeholder")

    print ("  Building a deep network with the following hidden layer structure (node numbers):")
    print ("\t\t", layers)
    print ("  for ", n_input, " input features and ", n_classes, " output classes")
    print ("  The following feature space variables are used for this network:")
    for col in Features:
        print ("\t", col)
    print ("\tWill set the network to have a single output layer")
    print ("""

    ---------------------------------------------------------
    The covariance matrix between these features is given by
    ---------------------------------------------------------

  """)
    data = sess.run(FeatureSpace)
    covMat = CalcCovarianceMatrix(data, Features)
    PrintMatrix(covMat)
    print ("""

    ---------------------------------------------------------
    The correlation matrix between these features is given by
    ---------------------------------------------------------

  """)
    corrMat = CalcCorrelationMatrix(data, Features)
    PrintMatrix(corrMat, "tex")

    print ("""

    ---------------------------------------------------------
    The means and RMS values of the features are given below
    ---------------------------------------------------------

  """)
    CalcMeansAndRMS(data, Features)


    output_layer = DNN(x, n_input, layers, options.n_output, options.DropOutKeepProb, "DNN", options.ActivationFunction)

#  cost = GetCost(output_layer, y, "softmaxcrossentropy")
# take an L2 loss function
    cost = tf.reduce_sum(tf.square(output_layer - y))
#  cost = GetLoss(output_layer, y, "L2")

    TryWeights = options.UseWeights
    # cost function; using cross entropy
    if(TryWeights):
        cost = GetCostWithWeight(output_layer, y, weight, "softmaxcrossentropy")

    # optimizer: take the Adam optimiser, see https://arxiv.org/pdf/1412.6980v8.pdf for
    # details of this algorithm.
    optimizer = GetOptimiser(cost, options.learning_rate, "adam")

    correct_prediction = tf.equal(tf.argmax(output_layer,1), tf.argmax(y,1))
    accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

    # Variables to fill during the training cycle in order to be able to make the summary plot
    avg_set = []
    epoch_set=[]
    train_accuracy = []

    print ("""
    --------------------------------------------------------
    Deep network training parameters are as follows:
    --------------------------------------------------------

  """)
    print ("\tDimensionality of feature space = ", n_input)
    print ("\tNumber of example types         = ", n_classes)
    print ("\tSum of example weights          = ", sumOfWeights)
    options.Print()

    # Start the session to embark on the training cycle
    init = tf.global_variables_initializer()
    with tf.Session() as sess:
        sess.run(init)

        VERBOSE = 0
        if VERBOSE:
            print ("data        = ", sess.run(FeatureSpace))
            print ("Feature set = ", Features)
            print ("Weights     = ", sess.run(Weights))
            print ("Labels      = ", sess.run(Labels))
        #
        # Convert the data into numpy arrays to feed into the optimisation process. This is a crude network
        # training where all examples presented are trained at each epoch.
        #
        batch_xs = sess.run(FeatureSpace)
        batch_ys = []
#      if(options.n_output == 1):
#          batch_ys = sess.run(Labels)
#      else:
        batch_ys = sess.run(n_classesTensor)
        batch_ws = sess.run(Weights) #sess.run(n_weightsTensor)

        # Training cycle - train all data; do not to batch training
        for epoch in range(options.training_epochs):
            avg_cost = 0.

            # Fit training using all data; train in batches to speed up optimisation step, but use all data
            n_batches = len(batch_xs)/options.batch_size
            # override size for negative batch sizes; those use all of the data
            if((options.batch_size < 0) or (options.batch_size > options.NTrainingData)):
                n_batches = 1
                options.batch_size = options.NTrainingData
                print ("Overriding MLPTrainingOptions settings. Will train using ", options.batch_size, " taining examples")

            for batch in range(n_batches):
                N_start = batch*options.batch_size
                N_end   = N_start + options.batch_size
                if(TryWeights):
                    sess.run(optimizer, feed_dict={x: batch_xs[N_start:N_end], y: batch_ys[N_start:N_end], weight: batch_ws[N_start:N_end]})
                else:
                    sess.run(optimizer, feed_dict={x: batch_xs[N_start:N_end], y: batch_ys[N_start:N_end]})
                # Compute average loss
                if(TryWeights):
                    avg_cost += sess.run(cost, feed_dict={x: batch_xs[N_start:N_end], y: batch_ys[N_start:N_end], weight: batch_ws[N_start:N_end]})/n_batches
                else:
                    avg_cost += sess.run(cost, feed_dict={x: batch_xs[N_start:N_end], y: batch_ys[N_start:N_end]})/n_batches

            # Display progress for each epoch; and store cost per epoch
            if epoch % options.display_step == 0:
                print ("Epoch:", '%04d' % (epoch+1), "cost=", "{:.9f}".format(avg_cost))
                avg_set.append(avg_cost)
                epoch_set.append(epoch+1)
                train_accuracy.append( accuracy.eval(feed_dict={x: batch_xs, y: batch_ys, keep_prob: options.DropOutKeepProb}) )

        if(options.Verbose):
            prediction=tf.argmax(output_layer,1)
            print ("predictions", prediction.eval(feed_dict={x: batch_xs}, session=sess))

            probabilities=output_layer
            print ("probabilities", probabilities.eval(feed_dict={x: batch_xs}, session=sess))

        # print summary
        print ("Training phase finished")
        if(len(epoch_set) == len(avg_set)):
            plt.plot(epoch_set,avg_set, 'o', label='DNN Training phase')
            plt.ylabel('cost')
            plt.xlabel('epoch')
            plt.legend()
            print ("Training plots will be output to ", options.TrainingPlotDir)
            MakeDir(options.TrainingPlotDir)
            TrainingPlotName = options.TrainingPlotDir
            TrainingPlotName += "dnn_training.pdf"
            plt.savefig(TrainingPlotName)
            plt.close("all")

        if(len(epoch_set) == len(train_accuracy)):
            plt.plot(epoch_set, train_accuracy, 'o', label='DNN Training phase')
            plt.ylabel('accuracy')
            plt.xlabel('epoch')
            plt.legend()
            TrainingPlotName = options.TrainingPlotDir
            TrainingPlotName += "dnn_accuracy.pdf"
            plt.savefig(TrainingPlotName)

    #
    # Before returning from this function we should make some basic book keeping plots, including
    # MVA output for each label, and the training plot of cost vs epoch.
    #
        print ("do book keeping")

    # return the model, cost function and optimiser for use by calling function (e.g.
    # for computing the network configuration or storing it etc.
    return output_layer, cost, optimizer, x


"""
========================================================================
 Training helper functions.  These are interfaces to cost and loss variables
 and to optimisation algorithms.

 This hipster does not provide support for weighted event samples. The error on
 the classification of an event should be multiplied by the event weight
 assigned for a particular training, test or validation example. This follows
 the logic of defining the overall loss function to optimise as used by the
 TMVA MLP node (UpdateNetwork function).

========================================================================
"""
def ImageLabelInputQueue(image_label_pairs,
                         num_epochs=None, shuffle=False):
    """
    -----------------------------------------------------------------
    TC
    Creates a string input producter from a list of image label pairs.

    image_label_pairs - a list of image label pairs

    returns: a tf input queue
    -----------------------------------------------------------------
    """
    input_queue = tf.train.string_input_producer(image_label_pairs,
                                                 num_epochs=num_epochs,
                                                 shuffle=shuffle)
    return input_queue

def GetCost(model, label, type):
    """
    ----------------------------------------------------------------------
     Create the cost function specified for the model and return that
    ----------------------------------------------------------------------
    """
    cost = 0
    if(type == "crossentropy"):
        print ("\tCreate a cross entropy cost function")
        cost = tf.reduce_mean ( -tf.reduce_sum( label*tf.log(model), reduction_indices=1 ) )
    elif (type == "softmaxcrossentropy"):
        cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=model, labels=label, name="cost"))
    elif (type == "sparsesoftmaxcrossentropy"):
        cost = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(logits=model, labels=label, name="cost"))
    elif (type == "softmax"):
        cost = tf.reduce_mean ( tf.nn.softmax(logits=model, dim=len(label[0]), name = "cost") )
    elif (type == "logsoftmax"):
        cost = tf.reduce_mean ( tf.nn.log_softmax(logits=model, dim=len(label[0]), name = "cost") )
#  elif (type == "weightedcrossentropy"):
#    loss = tf.nn.weighted_cross_entropy_with_logits(targets=labels, logits=model, pos_weight=???, name="cost")
    else:
        print ("ERROR - unknown cost function ", cost, " will return a softmax cross entropy cost")
        cost = GetCost(model, label, "softmaxcrossentropy")

    print ("Built cost function of type = ", type)

    return cost



def GetCostWithWeight(model, label, weight, type):
    """
    ----------------------------------------------------------------------
     Create the cost function specified for the model and return that
    ----------------------------------------------------------------------
    """
    cost = 0
    if(type == "crossentropy"):
        print ("\tCreate a cross entropy cost function")
        cost = tf.reduce_mean ( -tf.reduce_sum( weight*label*tf.log(model), reduction_indices=1 ) )
    elif (type == "softmaxcrossentropy"):
        cost = tf.reduce_mean(weight*tf.nn.softmax_cross_entropy_with_logits(logits=model, labels=label, name="cost"))
    elif (type == "sparsesoftmaxcrossentropy"):
        cost = tf.reduce_mean(weight*tf.nn.sparse_softmax_cross_entropy_with_logits(logits=model, labels=label, name="cost"))
    elif (type == "softmax"):
        cost = tf.reduce_mean ( weight*tf.nn.softmax(logits=model, dim=len(label[0]), name = "cost") )
    elif (type == "logsoftmax"):
        cost = tf.reduce_mean ( weight*tf.nn.log_softmax(logits=model, dim=len(label[0]), name = "cost") )
#  elif (type == "weightedcrossentropy"):
#    loss = tf.nn.weighted_cross_entropy_with_logits(targets=labels, logits=model, pos_weight=???, name="cost")
    else:
        print ("ERROR - unknown cost function ", cost, " will return a softmax cross entropy cost")
        cost = GetCostWithWeight(model, label, "softmaxcrossentropy")

    print ("Built cost function that uses weight with type = ", type)

    return cost


def GetLoss(model, label, type):
    """
    ----------------------------------------------------------------------
     Create the loss function specified for the model and return that
    ----------------------------------------------------------------------
    """
    loss = 0
    if(type == "crossentropy"):
        print ("\tCreate a cross entropy loss function")
        loss = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=model, labels=label))
    elif(type == "L2"):
        loss = tf.nn.l2_loss(label, model)
    elif(type == "logPoisson"):
        loss = tf.nn.log_poisson_loss(label, tf.log(model))

    return loss


def GetOptimiser(cost, learning_rate, type):
    """
    ----------------------------------------------------------------------
     Create an optimiser for the model.  The following options are implemented:

       gradiend descent: Newtonian gradient descent algorithm
                             type = gradientdescent, gradient, gd

       Ada Delta:
                             type = adadeltaoptimiser, adadelta, ada

       Ada Grad:
                             type = adagradoptimiser, adagrad

       Ada Grad DA:
                             type = adagradDAoptimiser, adagradDA

       Momentum:
                             type = momentumoptimiser, momentum

       ADAM:             Adaptive learning rate gradient descent algorithm, see
                         the following paper for detils: arXiv:1412.6980
                             type = adamoptimiser, adam

       FTRL:
                             type = ftrloptimiser, ftrl

       Proximal Gradient Descent:
                             type = proximalgradientdescentoptimiser, proximalgradientdescent, proximalgd, proxgd

       Proximal AdaGrad:
                             type = proximaladagradoptimiser, proximaladagrad, proximalada, proxada

       RMS Prop:
                             type = rmspropoptimiser, rmsprop, rms

     If an unknown optimiser type is specified a warning is printed and a default optimiser
     is returned.  The default optimiser is an ADAM optimiser as this provides superior
     performance over a simple gradient descent approach.

     See the TensorFlow optimiser page for details of these optimisers:
       https://www.tensorflow.org/hipster_guides/python/train#Optimizers
    ----------------------------------------------------------------------
    """
    optimizer = 0
    if((type == "gradientdescent") or (type == "gradient") or (type == "gd")):
        print ("\tCreating a gradient descent optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.GradientDescentOptimizer(learning_rate).minimize(cost)
    elif((type == "adadelta") or (type == "adadeltaoptimiser") or (type == "ada")):
        print ("\tCreating an Ada Delta optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.AdadeltaOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "adagrad") or (type == "adagradoptimiser")):
        print ("\tCreating an Ada Grad optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.AdagradOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "adagradDA") or (type == "adagradDAoptimiser")):
        print ("\tCreating an Ada Grad DA optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.AdagradDAOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "momentum") or (type == "momentumoptimiser")):
        print ("\tCreating an momentum optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.MomentumOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "adam") or (type == "adamoptimiser")):
        print ("\tCreating an ADAM optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "ftrl") or (type == "ftrloptimiser")):
        print ("\tCreating an FTRL optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.FtrlOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "proximalgradientdescent") or (type == "proximalgradientdescentoptimiser") or (type=="proximalgd") or (type=="proxgd")):
        print ("\tCreating an FTRL optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.ProximalGradientDescentOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "proximaladagrad") or (type == "proximaladagradoptimiser") or (type=="proximalada") or (type=="proxada")):
        print ("\tCreating an FTRL optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.ProximalAdagradOptimizer(learning_rate=learning_rate).minimize(cost)
    elif((type == "rmspropoptimiser") or (type == "rmsprop") or (type=="rms")):
        print ("\tCreating an FTRL optimiser for the model; learning rate = ", learning_rate)
        optimizer = tf.train.ProximalAdagradOptimizer(learning_rate=learning_rate).minimize(cost)
    else:
        print ("\tWARNING - unknown optimiser type, will default to an ADAM optimiser")
        optimiser = GetOptimiser(cost, learning_rate, "adamoptimiser")

    return optimizer


"""
========================================================================
 Helper functions for plotting follow on from here.  These can be used to
 display and export image data, for example MNIST examples. Those functions
 that rely on a particular community data source have that source in the
 function name; the other functions are generic.
========================================================================
"""
def PlotMNISTImage(example, type="train", action="plot"):
    """
    ----------------------------------------------------------------------
     Find the MNIST image example number specified, from the type of data
      type    = "test", "train", or "validation"
      action  = "plot" or <filename to print>
    ----------------------------------------------------------------------
    """
    print ("Importing the MNIST data via input_data interface")
    mnist = GetMNISTData()

    if(type=="train"):
        data = mnist.train.images[example:example+1,:]
        label = mnist.train.labels[example:example+1,:]
    elif(type=="test"):
        data = mnist.test.images[example:example+1,:]
        label = mnist.test.labels[example:example+1,:]
    elif(type=="validation"):
        data = mnist.validation.images[example:example+1,:]
        label = mnist.validation.labels[example:example+1,:]
    else:
        print ("Unknown MNIST data type - nothing to plot; known types are:")
        print ("\ttest")
        print ("\ttrain")
        print ("\tvalidation")
        return

    # the MNIST image data are arrays of 28x28 pixels; so reshape here
    img_size_x = 28
    img_size_y = 28
    image = data[0].reshape([img_size_x,img_size_y])
    if(action=="plot"):
        PlotMNISTExample(image, example, label)
    else:
        PrintMNISTExample(image, example, label, action)



def PlotMNISTExample(data, example, label):
    """
    ----------------------------------------------------------------------
     plot the image provided to the screen. The arguments are as follows:

       data      - appropriately resized image
       example   - example number
       label     - label for example image

    ----------------------------------------------------------------------
    """
    plt.imshow(data, cmap=plt.get_cmap('gray_r'))
    plt.title('Example: %d  Label: %d' % (example, label[0].argmax(axis=0)))
    plt.show()

def PlotImage(data, vmin=0, vmax=250, cmap='gray'):
    """
    ----------------------------------------------------------------------
    TC
    plot the image provided to the screen. The arguments are as follows:

    data      - appropriately resized image
    Optional Colourmap/Intensity Settings - Defaults to [0,255] Greyscale.

    cmap      - allowed colour map for imshow
    https://matplotlib.org/hipster/colors_hipster.html#matplotlib.colors.Colormap

    vmin      - min intensity
    vmax      - max intensity

    ----------------------------------------------------------------------
    """
    if np.asarray(data).ndim == 2:
                #GreyScale
        plt.imshow(data, cmap=plt.get_cmap(cmap), vmin=vmin, vmax=vmax)
    elif np.asarray(data).ndim == 3:
        if np.asarray(data).shape[2] == 1:
                #GreyScale
            x_dim = np.asarray(data).shape[0]
            y_dim = np.asarray(data).shape[1]
            resized_data = np.asarray(data).reshape(x_dim, y_dim)
            plt.imshow(resized_data, cmap=plt.get_cmap(cmap), vmin=vmin, vmax=vmax)
        elif np.asarray(data).shape[2] == 3:
            #RGB
            plt.imshow(data, cmap=plt.get_cmap(cmap), vmin=vmin, vmax=vmax)
        elif np.asarray(data).shape[2] == 4:
            #RGBA
            plt.imshow(data, cmap=plt.get_cmap(cmap), vmin=vmin, vmax=vmax)
    else:
        return("Invalid image size")


    plt.show()

def SaveArrayImage(name, array, multiple=False, ext='png', path=""):
    """
    ----------------------------------------------------------------------
    TC
    Save the array or arrays provided to disk (pixels saved as unsigned
    8bit ints).

    Arguments:
               
    name     -   The name of the file to save optionally prepended by a
                 path (in which case don't use the path variable). If
                 multiple is set to True, name should be a list of names.

    array    -   The np.array to be saved as an image. Array should be a
                 list of arrays if multiple is set to True.

    multiple -   Set True to save multiple images at once, otherwise False.

    ext      -   The file extension to save the image as, must be supported
                 by PIL. Default is png.

    path     -   Optionally specify a path here instead of in name, useful
                 if multiple=True then path is the directory and name can
                 just be a list of unique filenames.
    ----------------------------------------------------------------------
    """
    
    if not path == "":
        if not path[-1] == '/':
            path += '/'
            message = "to disk " + path
    else:
        message = "to disk here"
        
    if not multiple:
        output = Image.fromarray(array.astype(np.uint8))
        if not ext[0] == '.':
            output.save(path + name + '.' + ext)
            print(colours.INFO + "saving", name, message)
        else:
            output.save(path + name + ext)
            print(colours.INFO + "saving", name, message)
    elif multiple:
        if not len(name) == len(array):
            print(colours.ERROR + 'Length of array and name do not match (multiple is set to True)')
            raise ValueError('length of name does not match length of array')
        for i in len(name):
            output = Image.fromarray(array[i].astype(np.uint8))
            if not ext[0] == '.':
                output.save(name[i] + '.' + ext)
                print(colours.INFO + "saving", name[i], message)
            else:
                output.save(name[i] + ext)
                print(colours.INFO + "saving", name[i], message)
    else:
        print(colours.ERROR + 'multiple must be set to True or False')
        raise ValueError("multiple has an invalid value")

def PlotDataExample(data, example, label, shape):
    """
    ----------------------------------------------------------------------
     plot the image provided to the screen. The arguments are as follows:

       data      - appropriately resized image
       example   - example number
       label     - label for example image
       shape     - the shape of the image; used to resize the data

    ----------------------------------------------------------------------
    """
    image = data[0].reshape(shape)
    plt.imshow(image, cmap=plt.get_cmap('gray_r'))
    plt.title('Example: %d  Label: %d' % (example, label[0].argmax(axis=0)))
    plt.show()


def PlotAllFeatures(sess, TheData, columns, thisDir):
    """
    ----------------------------------------------------------------------
     Loop over the feature space and plot each of the features
    ----------------------------------------------------------------------
    """
    MakeDir(thisDir)

    for column in columns:
        thisFile = thisDir
        thisFile += '/'
        thisFile += column
        thisFile += '.pdf'
        PlotFeature(sess, TheData, columns, column, thisFile)


def PlotFeature(sess, TheData, columns, column, thisFile):
    """
    ----------------------------------------------------------------------
     Plot the specified feature as a histogram and output this to
     a file.
    ----------------------------------------------------------------------
    """
    print ("Creating a plot for ", column, ", will be written to the file", thisFile)

    # determine feature index
    featureIndex = 0;
    counter = 0;
    for col in columns:
        if col == column:
            featureIndex = counter
        counter+=1

    # extract the slice of data that corresponds to the column of interest
    thisdata = sess.run(TheData)
    featureToPlot = []

    for example in thisdata:
        # extract feature to plot
        featureToPlot.append(example[featureIndex])

    # construct the histogram and plot - this uses dummy data until the slicing part is
    # sorted out.
    fig = plt.figure()

    ax = fig.add_subplot(1, 1, 1)
    n, bins, patches = ax.hist(featureToPlot, 50)
    ax.set_xlabel(column)
    ax.set_ylabel("Number of entries")
    fig.savefig(thisFile)

    # clean up figure before returning
    plt.close(fig)

    return

def PlotAllScatterPlots(sess, TheData, columns, thisDir):
    """
    ----------------------------------------------------------------------
     Loop over the feature space and plot each non-trivial pair of features
    ----------------------------------------------------------------------
    """
    MakeDir(thisDir)

    for i in range(len(columns)):
        for j in range(len(columns)):
            if (i > j):
                thisFile = thisDir
                thisFile += '/'
                thisFile += columns[i]
                thisFile += '_'
                thisFile += columns[j]
                thisFile += '.pdf'
                ScatterPlot(sess, TheData, columns, columns[i], columns[j], thisFile)


def ScatterPlot(sess, TheData, columns, xfeat, yfeat, thisFile):
    """
    ----------------------------------------------------------------------
     Plot the specified features as xfeat and yfeat histogram and output this to
     a file.
    ----------------------------------------------------------------------
    """
    print ("Creating a plot for ", xfeat, " vs ", yfeat, ", will be written to the file", thisFile)

    # determine feature index
    xfeatureIndex = 0;
    yfeatureIndex = 0;
    counter = 0;
    for col in columns:
        if col == xfeat:
            xfeatureIndex = counter
        if col == yfeat:
            yfeatureIndex = counter
        counter+=1

    # extract the slice of data that corresponds to the column of interest
    thisdata = sess.run(TheData)
    xfeatureToPlot = []
    yfeatureToPlot = []

    for example in thisdata:
        # extract feature to plot
        xfeatureToPlot.append(example[xfeatureIndex])
        yfeatureToPlot.append(example[yfeatureIndex])

    # construct the histogram and plot - this uses dummy data until the slicing part is
    # sorted out.
    fig = plt.figure()

    ax = fig.add_subplot(1, 1, 1)
    ax.scatter(xfeatureToPlot[:], yfeatureToPlot[:], c='b')
    ax.set_xlabel(xfeat)
    ax.set_ylabel(yfeat)

    fig.savefig(thisFile)

    # clean up figure before returning
    plt.close(fig)

    return

def PrintMNISTExample(data, example, label, fname):
    """
    ----------------------------------------------------------------------
     print the image provided to the specified file.  The arguments are as follows:

       data      - appropriately resized image
       example   - example number
       label     - label for example image
       fname     - output file name

    ----------------------------------------------------------------------
    """
    plt.imshow(data, cmap=plt.get_cmap('gray_r'))
    plt.title('Example: %d  Label: %d' % (example, label[0].argmax(axis=0)))
    plt.savefig(fname)


def PrintDataExample(data, example, label, shape, fname):
    """
    ----------------------------------------------------------------------
     plot the image provided to the screen. The arguments are as follows:

       data      - appropriately resized image
       example   - example number
       label     - label for example image
       shape     - the shape of the image; used to resize the data
       fname     - output file name

    ----------------------------------------------------------------------
    """
    image = data[0].reshape(shape)
    plt.imshow(image, cmap=plt.get_cmap('gray_r'))
    plt.title('Example: %d  Label: %d' % (example, label[0].argmax(axis=0)))
    plt.savefig(fname)



"""
========================================================================
 Helper functions for calculations follow on from here.
========================================================================
"""
def CalcCorrelationMatrix(Data, columns):
    """
    ----------------------------------------------------------------------
     Compute the correlation matrix for all combinations of pairs of variables

     Data       - this is a python array
     columns    - this is a python list of column headings
    ----------------------------------------------------------------------
    """
    corrMatrix=[]#create an empty list first
    for i in range(len(columns)):
        corrMatrix.append([0]*len(columns))
        for j in range(len(columns)):
            corrMatrix[i][j] = 0.0

    for i in range(len(columns)):
        for j in range(len(columns)):
            if(i>=j):
                dummy, corrMatrix[i][j] = CalcCovariance(Data, i, j)

    for i in range(len(columns)):
        for j in range(len(columns)):
            if(i>j):
                corrMatrix[j][i] = corrMatrix[i][j]

    return corrMatrix



def CalcCorrelationMatrixT(sess, DataTensor, columns):
    """
    ----------------------------------------------------------------------
     Compute the correlation matrix for all combinations of pairs of variables

     DataTensor - this is a tensor
     columns    - this is a python list of column headings
    ----------------------------------------------------------------------
    """
    Data = sess.run(DataTensor)
    correlationMatrix = CalcCorrelationMatrix(Data, columns)

    return correlationMatrix




def CalcCovariance(Data, col1, col2):
    """
    ----------------------------------------------------------------------
     Compute the covariance (and correlation) between tensor columns col1 and col2

     Data       - this is a python array
     col1       - the first index (this is an integer)
     col2       - the second index (this is an integer)
    ----------------------------------------------------------------------
    """

    # need to compute the mean
    MeanCol1 = 0
    MeanCol2 = 0
    for i in range(len(Data)):
        ThisExample = Data[i]
        MeanCol1 += ThisExample[col1]
        MeanCol2 += ThisExample[col2]

    MeanCol1 = MeanCol1/len(Data)
    MeanCol2 = MeanCol2/len(Data)

    # now compute the sum of the residuals (squared)
    sResCol1 = 0
    sResCol2 = 0
    sResProd = 0
    for i in range(len(Data)):
        ThisExample = Data[i]
        ResCol1 = (ThisExample[col1] - MeanCol1)
        ResCol2 = (ThisExample[col2] - MeanCol2)

        sResCol1 += ResCol1*ResCol1
        sResCol2 += ResCol2*ResCol2
        sResProd += ResCol1*ResCol2

    # take the residual sums (squared) and convert those to variances
    sResCol1 = sResCol1/(len(Data)-1)
    sResCol2 = sResCol2/(len(Data)-1)
    sResProd = sResProd/(len(Data)-1)

    covar = sResProd
    corr  = sResProd/math.sqrt(sResCol1)/math.sqrt(sResCol2)

    return covar, corr



def CalcCovarianceT(sess, DataTensor, col1, col2):
    """
    ----------------------------------------------------------------------
     Compute the covariance (and correlation) between tensor columns col1 and col2

     DataTensor - this is a tensor
     col1       - the first index (this is an integer)
     col2       - the second index (this is an integer)
    ----------------------------------------------------------------------
    """
    Data = sess.run(DataTensor)
    covariance = CalcCovariance(Data, col1, col2)

    return covariance



def CalcCovarianceMatrix(Data, columns):
    """
    ----------------------------------------------------------------------
     Compute the covariance matrix for all combinations of pairs of variables

     Data - this is a python array
     this is a python list of column headings
    ----------------------------------------------------------------------
    """
    covMatrix=[]#create an empty list first
    for i in range(len(columns)):
        covMatrix.append([0]*len(columns))
        for j in range(len(columns)):
            covMatrix[i][j] = 0.0

    for i in range(len(columns)):
        for j in range(len(columns)):
            if(i>=j):
                covMatrix[i][j], dummy = CalcCovariance(Data, i, j)

    for i in range(len(columns)):
        for j in range(len(columns)):
            if(i>j):
                covMatrix[j][i] = covMatrix[i][j]

    return covMatrix


def CalcCovarianceMatrixT(sess, DataTensor, columns):
    """
    ----------------------------------------------------------------------
     Compute the covariance matrix for all combinations of pairs of variables

     DataTensor - this is a tensor
     this is a python list of column headings
    ----------------------------------------------------------------------
    """
    Data = sess.run(DataTensor)
    covarianceMatrix = CalcCovarianceMatrix(Data, columns)

    return covarianceMatrix

def CalcMeansAndRMS(Data, columns):
    """
    ----------------------------------------------------------------------
     Compute the mean and RMS values for all columns

     Data       - this is a python array
     columns    - this is a python list of column headings
    ----------------------------------------------------------------------
    """
    Ndata = len(Data)
    thisMean = 0
    thisSum  = 0
    for j in range(Ndata):
        thisMean += Data[j]
    thisMean = thisMean / Ndata

    for j in range(Ndata):
        thisSum += (Data[j]-thisMean)*(Data[j]-thisMean)
    thisRMS = thisMean / math.sqrt(Ndata-1.0)

    for i in range(len(columns)):
        max, min = GetFeatureMaxMin(Data, i)
        print ("\t mean = ", thisMean[i], "\t RMS = ", thisRMS[i], "\t ", columns[i], " in [", min, ", ", max, "]")


    return

def PrintMatrix(matrix, format="normal"):
    """
    ----------------------------------------------------------------------
     Function to print out a covariance/correlation matrix for a 2D array

     matrix          - a python array of the type produced by CalcCovarianceMatrix
                       of CalcCorrelationMatrix
     format          - A string specifying output format:
                           normal  : standard text printout to stdio
                           tex     : LaTeX formatted table printout to stdio
    ----------------------------------------------------------------------
    """
    if(format == "normal"):
        for i in range(len(matrix)):
            string = "\t"
            for j in range(len(matrix)):
                string += "{:.2f}".format(matrix[i][j])
                string += " "
            print (string)
    elif((format == "tex") or (format == "latex")):
        print ("\\begin{table}[!ht]")
        print ("\\caption{Matrix caption to complete}\n\\label{tbl:dummy}")
        printstring = "\\begin{tabular}{"
        for i in range(len(matrix)):
            printstring += "l"
        printstring += "}\n\\hline\\hline"
        print (printstring)
        for i in range(len(matrix)):
            string = "\t"
            for j in range(len(matrix)):
                string += "{:.2f}".format(matrix[i][j])
                string += " & "
            string += "\\\\"
            print (string)
        print ("\\hline\\hline\n\\end{tabular}\n\\end{table}")
    return


"""
========================================================================
 Helper functions for data I/O follow on from here.  These include
 CSV IO, where the file format is the first row providing a comma
 separated list of variables (the columns), and subsequent rows are
 data examples.

 e.g.
       xdata = hipster.ExtractColumnTensor(sess, dataTensor, columns, 'x')

 where sess is the session dataTensor is a TensorFlow variable containing
 the data, columns is a list of column names and 'x' is the column name
 of interest, passed as a string.
========================================================================
"""
def ArraySplit(array, index=0, partitions=4):
    """
    TC
    Splits an array into a number of partitions.
    
    Arguments:
        array      - the array you want to split
        index      - the index to split on
        partitions - the number of partitions to split into

    Returns:
        arrays     - a list of arrays
    """
    size = array.shape[index]
    per_partition = int(size/partitions)
    arrays = []
    i_1 = 0
    i_2 = per_partition
    for _ in range(partitions):
        arrays.append(array[i_1:i_2])
        i_1 += per_partition
        i_2 += per_partition

    return arrays

def ReadImage(filename_list):
    """
    ----------------------------------------------------------------------
     TC
     Read image files from a directory.

     Arguments:
       filename_list: a list of paths to images, or a single path

     Output:
       arrays: a list of images stored as numpy arrays

    N.B. if filename_list is just a string one array will be returned
    (not in a list)
    ----------------------------------------------------------------------
    """
    if isinstance(filename_list, str):
        image = Image.open(filename_list)
        return image
    else:
        arrays = []
        for f in filename_list:
            image = Image.open(f)
            arrays.append(image)
        return arrays



def WriteNPY(directory, arrays):
    """
    ----------------------------------------------------------------------
     KF
     Write arrays into .npy files. Each file contains one numpy array.

     Arguments:
       directory: output folder
       arrays: list of arrays to be written into files

     Name of the resulting files will be imageX.npy,
     where X is the index starting from 0.
    ----------------------------------------------------------------------
    """
    hipster.MakeDir(directory)
    writing_epoch = 0
    for i in arrays:
        np.save(directory + 'image%d.npy' % writing_epoch, arrays[writing_epoch])
        writing_epoch += 1



def ReadNPY(filename_list):
    """
    ----------------------------------------------------------------------
     KF
     Read arrays from .npy files.

     Arguments:
       filename_list: a list of paths to .png images (explicit, no wildcard support)

     Output:
       readout: a list of images stored as numpy arrays
    ----------------------------------------------------------------------
    """
    readout = []
    for i in filename_list:
        readout.append(np.load(i))
    return readout


def CutOnData(sess, TFVdata, columns, columnToCut, min, max):
    """
    ----------------------------------------------------------------------
     Apply a cut on the data for the specified column, retain values between
     the minimum and maximum numerical values specified.
    ----------------------------------------------------------------------
    """

    # get the column index
    thisColIndex = -999
    for i in range(len(columns)):
        if columnToCut == columns[i]:
            thisColIndex = i
            break

    if thisColIndex == -999:
        print ("ERROR - Unable to locate the column ", columnToCut, " in order to place a cut on the data")
        return TFVdata

    thisData = sess.run(TFVdata)
    newData = []
    for i in range(len(thisData)):
        thisExample = thisData[i]
        if (thisExample[thisColIndex] < max) and (thisExample[thisColIndex] > min):
            thisRow = []
            for j in range(len(thisExample)):
                thisRow.append( thisExample[j] )
            newData.append(thisRow)

    newDataTensor = tf.Variable(newData, name = "reduced_data")

    return newDataTensor


def CutOutNegativeData(sess, TFVdata, columns, columnToCut):
    """
    ----------------------------------------------------------------------
     Apply a cut on the data for the specified column, retain values that
     are >= 0, and reject all negative values.  For example this helper
     function can be used to remove examples with negative weights prior
     to plotting/training etc.
    ----------------------------------------------------------------------
    """
    # get the column index
    thisColIndex = -999
    for i in range(len(columns)):
        if columnToCut == columns[i]:
            thisColIndex = i
            break

    if thisColIndex == -999:
        print ("ERROR - Unable to locate the column ", columnToCut, " in order to remove examples with negative values")
        return TFVdata

    init = tf.global_variables_initializer()
    sess.run(init)
    thisData = sess.run(TFVdata)
    newData = []

    for i in range(len(thisData)):
        thisExample = thisData[i]
        if (thisExample[thisColIndex] >= 0):
            thisRow = []
            for j in range(len(thisExample)):
                thisRow.append( thisExample[j] )
            newData.append(thisRow)

    newDataTensor = tf.Variable(newData, name = "reduced_data")

    return newDataTensor




def ExtractColumnTensor(sess, TFVdata, columns, column):
    """
    ----------------------------------------------------------------------
     create a tensor corresponding to a single column in the data. The inputs
     are the session (sess), a data tensor (TFVdata), a list of columns of
     interest (columns) and a string of the column to extract (column).

     This function creates a new tensor, and the tensor needs to be initialised
     before it can be used.
    ----------------------------------------------------------------------
    """
    newdata = ExtractColumn(sess, TFVdata, columns, column)
    thisname = "column_data_"
    thisname += str(column)
    data = tf.Variable( newdata, name=thisname )

    return data


def ExtractColumnsTensor(sess, TFVdata, columns, subcolumns):
    """
    ----------------------------------------------------------------------
     create a tensor corresponding to specified columns of the data. The inputs
     are the session (sess), a data tensor (TFVdata), a list of columns of
     interest (columns) and a list of the columns to extract (subcolumns).

     This function creates a new tensor, and the tensor needs to be initialised
     before it can be used.
    ----------------------------------------------------------------------
    """
    newdata = ExtractColumns(sess, TFVdata, columns, subcolumns)
    thisname = "column_data_"
    for col in subcolumns:
        thisname += str(col)
        thisname += "_"

    data = tf.Variable( newdata, name=thisname )

    return data



def ExtractColumn(sess, TFVdata, columns, column):
    """
    ----------------------------------------------------------------------
     extract a column of data from the tensor; using the column name to identify
     the index of the tensor data to extract.

        sess        - the session
        TFVdata     - A TensorFlow tensor of data with dimensions [N, dim(x)]
        columns     - the column names
        column      - the name of the column to extract from the data

    ----------------------------------------------------------------------
    """
    columnstring = ""
    counter = 0
    colNumber = -1
    thisColumn = []
    for col in columns:
        if(column == col):
            colNumber = counter
            break
        counter += 1

    if colNumber < 0:
        print ("ERROR - unable to extract column ", column, " from the data sample; returning a null list")
        return thisColumn

    thisData = sess.run(TFVdata)
    N = len(thisData)
    print ("Extracting column ", colNumber, " (", column, ") from the data. There are N = ", N, " examples in this data set")

    for i in range( N ):
        thisColumn.append( thisData[i][colNumber]  )

    return thisColumn



def ExtractColumns(sess, TFVdata, columns, subcolumns):
    """
    ----------------------------------------------------------------------
     extract a subspace of the input feature space from the data
     the index of the tensor data to extract.

        sess        - the session
        TFVdata     - A TensorFlow tensor of data with dimensions [N, dim(x)]
        columns     - the column names
        subcolumns  - the name of the columns to extract from the data

    ----------------------------------------------------------------------
    """
    # loop over the subcolumns to find the correspinding indices that match the columns.
    colNumbers = []
    for i in range(len(subcolumns)):
        counter = 0
        for col in columns:
            if(subcolumns[i] == col):
                colNumbers.append(counter)
                break
            counter += 1

    # now extract the data in the appropriate ordering
    dataToProcess = sess.run(TFVdata)
    N = len(dataToProcess)
    print ("Extracting column ", colNumbers, " from the data. There are N = ", N, " examples in this data set")
    print ("The extracted columns correspond to the feature labels ", subcolumns)

# Convert the tensor back to an array so only run the session once; otherwise the
# extraction of columns is inefficient

    thisDataSubset = []
    for i in range( N ):
        # create a copy of the example data that will only contain the examples in the
        # order presented by the subcolumns variable.
        thisExample = []
        for j in range( len(subcolumns) ):
            thisExample.append( dataToProcess[i][ colNumbers[j] ] )

        thisDataSubset.append( thisExample )
        if not ((i+1) % 50000):
            print ("\tProcessed ", i+1, " examples")

    return thisDataSubset


def MergeFeatureSets(sFeatureSpace, bFeatureSpace, sLabels, bLabels, sWeights, bWeights, NtoMerge=-999, NormalisationType="[0,1]"):
    """
    ----------------------------------------------------------------------
     Given input data sets corresponding to signal and background examples,
     construct an ensemble sample that interleaves the signal and background
     training examples; ensure that labels and weights are adequately
     tracked into the new sample.

      sFeatureSpace  - signal feature space
      bFeatureSpace  - background feature space
      sLabels        - signal labels
      bLabels        - background labels
      sWeights       - signal weights
      bWeights       - background weights
      NtoMerge       - the number of events to merge; if -999 then use the largest possible numbers
                       given the sample sizes, where equal example numbers are retained.
      NormalisationType
                     - default [0,1]: map features onto this range
                     - none: do not map features
                     - [-1,1]: map features onto this range

     This function will return three variables; the data, weights and
     labels (all prepared as tensors)
    ----------------------------------------------------------------------
    """
    sess = tf.Session()
    init = tf.global_variables_initializer()
    sess.run(init)

    sDat = sess.run(sFeatureSpace)
    bDat = sess.run(bFeatureSpace)
    sL   = sess.run(sLabels)
    bL   = sess.run(bLabels)
    sW   = sess.run(sWeights)
    bW   = sess.run(bWeights)

    Nsig = len(sDat)
    Nbg  = len(bDat)
    N = Nsig
    if(Nsig > Nbg):
        N = Nbg
    if (NtoMerge >0) and (NtoMerge < 2.0*N):
        N = int(NtoMerge/2.0)

    print ("MergeFeatureSets: Will construct a merged sample using ", N, " examples")
    print ("                  from the signal and background data sets.")
    # use the smallest number of examples as the number to use from either sample.
    newData    = []
    newLabels  = []
    newWeights = []
    for i in range(N):
        # do element-wise copy of the data for tensor construction
        sigRow = []
        bgRow = []
        for val in sDat[i]:
            sigRow.append(val)
        for val in bDat[i]:
            bgRow.append(val)

        newData.append( sigRow )
        newData.append( bgRow )

        newLabels.append( sL[i] )
        newLabels.append( bL[i] )

        newWeights.append( sW[i] )
        newWeights.append( bW[i] )

    print ("Created a merged data sample with ", len(newData), " examples from signal and background")
    newDataTensor    = tf.Variable(newData,    name = "merged_data")
    newLabelsTensor  = tf.Variable(newLabels,  name = "merged_labels")
    newWeightsTensor = tf.Variable(newWeights, name = "merged_weights")

    # check to see if we need to transform the data or not.  By deafault the feature
    # space is normalised onto the coordinate range [0,1] for all variables.
    # The following mapping options are:
    #    [0,1]            <---- default behaviour
    #    [-1,1]
    #    none
    init = tf.global_variables_initializer()
    sess.run(init)
    newData = sess.run(newDataTensor)
    if(NormalisationType == "[0,1]"):
        thisData = NormaliseData(newData)
        newDataTensor = tf.Variable(thisData, "normalised_merged_data")
        print ("Merged data sample has been created - feature space is mapped into the range [0, 1]")
    elif (NormalisationType == "[-1,1]"):
        thisData = NormaliseDataNeg1To1(newData)
        newDataTensor = tf.Variable(thisData, "normalised_merged_data")
        print ("Merged data sample has been created - feature space is mapped into the range [-1, 1]")
    elif (NormalisationType == "none"):
        print ("Merged data sample has been created - feature space is not mapped")
    init = tf.global_variables_initializer()
    sess.run(init)

    print ("Merged data tensor shape is:")
    print (newDataTensor)

    return newDataTensor, newLabelsTensor, newWeightsTensor



def PrintDataTFV(sess, TFVdata, columns, option=""):
    """
    ----------------------------------------------------------------------
     Given a data sample with columns specified, print the data in a convenient format
        sess       - the TensorFlow session
        TFVdata    - data in the form of a TensorFlow Variable
        columns    - column names in the form of a python variable
        option     - print the first 10 examples, unless option="all"
    ----------------------------------------------------------------------
    """
    # now convert the data back to a python array and print it out
    PyData = sess.run(TFVdata)

    # first prepare the column names to print out, this is not a csv file
    columnstring = ""
    for col in columns:
        columnstring += str(col)
        columnstring += " "
    print ("There are N = ", len(PyData), " examples in this data set; printed out below")
    print (columnstring)


    N = len(PyData)
    if not (option=="all"):
        if N > 10:
            N = 10

    for i in range(N):
        #loop over events and print output
        example = PyData[i]
        thisLine = ""
        for i in range(len(example)):
            thisLine += str(example[i])
            thisLine += " "
        print (thisLine)

    return




def ReadCSVasTFV(filename, option=""):
    """
    ----------------------------------------------------------------------
     read the CSV file as a tensor flow variable, use ReadCSV to import
     the data as an array, and then convert to a TF variable. Requires a
     session to be running.

     To use this function you first call the function, and then need to
     start the session and initilise the graph using, for example:

         data, columns, stdata, stcolumns = hipster.ReadCSVasTFV("testdata.csv")

         sess = tf.Session()
         init = tf.global_variables_initializer()
         sess.run(init)

         # no the tf variable can be used, this will have shape [Ndata, dim(x)]
         print sess.run(data3tf)

     This function returns a TensorFlow Variable for the data set and a
     python list for the column names.
    ----------------------------------------------------------------------
    """
    data, columns, stringdata, stringcolumns = ReadCSV(filename, option)
    dataTensor = tf.Variable(data, name = "data_from_csv_file")
    stringDataTensor = tf.Variable(data, name = "string_data_from_csv_file")

    return dataTensor, columns, stringDataTensor, stringcolumns



def ReadCSV(filename, option=""):
    """
    ----------------------------------------------------------------------
     Read in the specified CSV file; the first line is a comma separated list of
     columns to be read into the CSV file. The dimensionality of that sets the
     dimensionality of the data.  If events do not match the column heading
     dimensionality then an error is printed out and the event is skipped.

     This function returns:
       theData       : a list of lists / the data (floating point representation)
       theDataColumns: a list of column names
       theStringData : a list of lists of data for string based information
                       (e.g. string labels)
       theStringDataColumns: a list of column names
    ----------------------------------------------------------------------
    """
    print ("Will read the file ", filename)

    # strip the newline character from the end of the line, and then split the data to create columns
    data    = open(filename).readlines()
    data[0] = data[0].strip()
    columns = data[0].split(',')

    VERBOSE = bool(0)
    QUIET   = bool(0)
    VERBOSE, QUIET = parseOptionsIO(option, VERBOSE, QUIET)

    if VERBOSE:
        print ("\tThis file contains the following columns")
        for column in columns:
            print ("\t\t", column)
        print ("\tand contains ", len(data)-1, " entries")
    elif not QUIET:
        print ("\tThis file contains ", len(columns), " columns and ", len(data)-1, " entries")

    # loop over the data, for each line split the data on commas,
    # suppress white space characters and convert to numbers
    # to store in an array for.  The variable theData is the
    # numerical data of interest, and will be returned to the
    # calling function.
    theData = []
    theStringData = []
    skippedEvents = 0
    theDataColumns = []
    theStringDataColumns = []
    for i in range(len(data)-1):
        thisRow = []
        thisStringRow = []
        thisLine = data[i+1]
        thisLine = thisLine.replace(" ", "")
        csvLineData = thisLine.split(',')
        if len(csvLineData) != len(columns):
            print ("\tERROR READING LINE ", i, " (wrong number of columns for this row, found ", len(csvLineData), " expected ", len(columns), " )")
            skippedEvents+=1
        else:
            for j in range(len(csvLineData)):
                try:
                    float(csvLineData[j])
                    thisRow.append( float(csvLineData[j]) )
                    if i == 0:
                        theDataColumns.append(columns[j])
                except ValueError:
                    thisStringRow.append( csvLineData[j] )
                    if i == 0:
                        theStringDataColumns.append(columns[j])
            if VERBOSE :
                print (thisRow)

            # append the data samples with lvalue and string features
            if(len(thisRow)):
                theData.append(thisRow)
            if(len(thisStringRow)):
                theStringData.append(thisStringRow)

    if(skippedEvents):
        print ("\tWARNING: a total of ", skippedEvents, " have been skipped when reading this data in - data file is not properly formatted")

    if not QUIET:
        print ("\tThe file has been read")


    return theData, theDataColumns, theStringData, theStringDataColumns





def WriteCSV(filename, data, vars, option=""):
    """
    ----------------------------------------------------------------------
     Write the specified CSV file; the first line is a comma separated list of
     columns to be read into the CSV file. The dimensionality of that sets the
     dimensionality of the data.

       filename is the name of the file to be created; existing files will be
                overwritten.
       data is the array to be written; each element (row) of the array will
                be converted into a comma separated string.
       vars is the list of variables to be written out (the column headings)

    ----------------------------------------------------------------------
    """
    VERBOSE = bool(0)
    QUIET   = bool(0)
    VERBOSE, QUIET = parseOptionsIO(option, VERBOSE, QUIET)

    if not QUIET:
        print ("Will write the file ", filename)

    if len(vars) != len(data[0]) :
        print ("\tERROR; can not write out CSV file ", filename, " as there are different data dimensions and variable column number")
        return

    # open the file for output
    f = open(filename, 'w+')

    # start the output by printing the column headings
    columnsToWrite = ""
    for i in range(len(vars)):
        columnsToWrite += vars[i]
        if(i < len(vars)-1):
            columnsToWrite += ","
    if VERBOSE:
        print (columnsToWrite)
    columnsToWrite += "\n"
    f.write(columnsToWrite)

    # now print the data line by line; need to insert commas and new line
    # characters when writing the data out in this form
    for i in range(len(data)):
        dataToWrite = ""
        thisData = data[i]
        length = len(thisData)
        for j in range(len(thisData)):
            dataToWrite += str(thisData[j])
            if j < length - 1:
                dataToWrite += ","

        if VERBOSE:
            print (dataToWrite)
        dataToWrite += "\n"

        f.write(dataToWrite)

    return



def WriteTFVasCSV(sess, filename, TFdata, vars, option=""):
    """
    ----------------------------------------------------------------------
     write the tf variable of a data sample as a CSV file.  Use the WriteCSV
     function to handle output so that this function only deals with converting
     the tf variable to a python array with dimension [Ndata, dim(x)] that
     can then be output to the file filename.

         print sess.run(data3tf)
    ----------------------------------------------------------------------
    """
    data = sess.run(TFdata)
    WriteCSV(filename, data, vars, option)

    return



def parseOptionsIO(option, VERBOSE, QUIET):
    """
    ----------------------------------------------------------------------
     this is a helper function for I/O options
    ----------------------------------------------------------------------
    """
    if(option == "v" or option == "verbose"):
        VERBOSE = 1
    if(option == "q" or option == "quiet"):
        QUIET = 1

    return VERBOSE, QUIET


def GetMNISTData(directory="/tmp/data/"):
    print ("""
  Attempting to load the MNIST data; this will be written to /tmp/data.
  """)
    mnist = read_data_sets(directory, one_hot=True)
    return mnist

def MakeDir(outputdir):
    """
    ----------------------------------------------------------------------
     If a directory does not exist, make it.  If this fails then ensure
     that an error is clearly reported to the user and return an error code
    ----------------------------------------------------------------------
    """
    directory = os.path.dirname(outputdir)
    if not os.path.exists(directory):
        print ("Directory ", outputdir, " does not exist - creating this")
        os.makedirs(directory)
        directory2 = os.path.dirname(outputdir)
        if not os.path.exists(directory2):
            print ("\tERROR - unable to create the directory: ", outputdir)
            return 2
    else:
        return 1

    return 0

def FilenameList(source, extension):
    """
    -----------------------------------------------------------------
     TC
     Walk through a source directory and generate a list of all files
     with specified extension. Also supports a list of extensions.

       source     - string: the path to the source directory
       extension  - string/list(string): selected extension/s

       returns: list of filenames in source with extension: extension
    -----------------------------------------------------------------
    """
    filename_list = []
    for root, dirnames, filenames in os.walk(source):
        for filename in filenames:
            if not isinstance(extension, str):
                for i in range(len(extension)):
                    if filename.endswith(extension[i]):
                        filename_list.append(os.path.join(root,
                                                          filename))
            else:
                if filename.endswith(extension):
                    filename_list.append(os.path.join(root,
                                                      filename))
    return filename_list

def DirectoryList(source):
    """
    -----------------------------------------------------------------
    TC
    Walk through a source directory and generate a list of all files
    with specified extension. Also supports a list of extensions.

    source     - string: the path to the source directory
    extension  - string/list(string): selected extension/s

    returns: list of filenames in source with extension: extension
    -----------------------------------------------------------------
    """

    directory_list = next(os.walk(source))[1]

    return directory_list

def ImageLabelMap(filename_list, label_list, delimitter=' ', file_path=None):
    """
    -----------------------------------------------------------------
    TC
    Takes a list of filenames and labels generating a .txt file
    in which each line represents 'example label'.

    filename_list  - list(string): a list of filenames
    label_list     - list(string): a list of labels

    ~  Optional ~
    delimitter     - the delimitter used in the .txt file.
    file_path      - path to .txt file, creates unique one if
                     unspecified

    returns: list of filenames in source with extension: extension
    raises: Error if filename_list and label_list have different lengths.
    -----------------------------------------------------------------
    """
    if not len(filename_list) == len(label_list):
        raise ValueError("List lengths incompatible.")

    if not file_path:
        unique_string = str(uuid.uuid4())
        f = open('Image_Label_Map_' + unique_string + '.txt', 'a')
    elif not os.path.isfile(file_path):
        f = open(file_path, 'w')
    else:
        print("INFO: appending to existing file map")
        f = open(file_path, 'a')
    for i in range(len(filename_list)):
        f.write(filename_list[i] + delimitter + str(label_list[i]) + '\n')

def ReadImageLabelMap(image_label_map):
    """
    -----------------------------------------------------------------
    TC
    Reads an Image/Filename map like the one created by ImageLabelMap.

    image_label_map - path to the map file (.txt usually).

    returns: a list of image label pairs [image0 label0, ..., imageN labelN]
    -----------------------------------------------------------------
    """
    f = open(image_label_map, 'r')
    image_label_pairs = []
    for line in f:
        image_label_pair = line
        image_label_pairs.append(image_label_pair.rstrip())
    return image_label_pairs

def ReadImagesFromDisk(image_and_label_tensor, file_type, delimitter=' '):
    """
    ------------------------------------------------------------------------------------------
     TC
     Reads labelled images from disk.

       image_and_label_tensor - tensor: a single image label pair for example a
                                single item in the queue created by ImageLabelInputQueue

       file_type              - string: specifies the file format of the images
       delimitter             - string: specifies how images and labels are separated

       returns: a single decoded image and its label
    ------------------------------------------------------------------------------------------
    """
    if delimitter == ' ':
        filename, label = tf.decode_csv(image_and_label_tensor,
                                    [[""], [""]], " ")
    else:
        filename, label = tf.decode_csv(image_and_label_tensor,
                                    [[""], [""]], delimitter)
    file_contents = tf.read_file(filename)
    if file_type == 'jpeg':
        example = tf.image.decode_jpeg(file_contents)
    elif file_type == 'png':
        example = tf.image.decode_png(file_contents)
    else:
        "File type not supported, please use jpeg or png."

    return example, label

def int64Feature(value):
    """
    -----------------------------------------------------------------
     TC
     Shorthand for an int64 tf.train.Feature
    -----------------------------------------------------------------
    """
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def bytesFeature(value):
    """
    -----------------------------------------------------------------
     TC
     Shorthand for an bytes tf.train.Feature
    -----------------------------------------------------------------
    """
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

"""
----------------------------------------------------------------------------------

               StackJpegs IS UNUSED: Might be useful at a later stage

----------------------------------------------------------------------------------
"""
def StackJpegs(paths_to_images):
    """
    ------------------------------------------------------------------------------------------
     TC
     Takes a list of paths to images and returns a stack of images.
     E.G. if images are size (width, height) a stack of N images has shape (width, height, N)
     and to get the nth slice (and therefore nths image in the stack) one would slice out
     (width, height, n)

       paths_to_images - list(string): paths to jpegs

       returns: a stacked tensor of all the images
    ------------------------------------------------------------------------------------------
    """
    images = []
    for i in range(len(paths_to_images)):
        file_contents = tf.read_file(paths_to_images[i])
        image = tf.image.decode_jpeg(file_contents)
        images.append(image)

    image_stack = tf.stack(images)
    return image_stack

def WriteTFRecord(sess, image_label_pairs, oname, opath, parent_type='jpeg', delimitter=' '):
    """
    ------------------------------------------------------------------------------------------
     TC
     Writes a TFRecord file encoding images and corresponding labels.
     !!! NB: This function is poorly named and will have to be revised at a later date. !!!

       image_label_pairs - list(string): a list of image label pairs [image0 label0, ..., imageN
                                         labelN]
       oname             - string: name of the output .tfrecords file
       opath             - string: path to directory where to save the .tfrecords file

       returns: None
       NB: Sess.run this.
    ------------------------------------------------------------------------------------------
    """
    labels = []
    paths_to_images = []
    if delimitter==' ':
        for i in range(len(image_label_pairs)):
            path, label = image_label_pairs[i].split(' ')
            labels.append(label)
            paths_to_images.append(path)
    else:
        for i in range(len(image_label_pairs)):
            path, label = image_label_pairs[i].split(delimitter)
            labels.append(label)
            paths_to_images.append(path)

    num_examples = len(image_label_pairs)
    filename = opath + oname + '.tfrecords'
    print('Writing', filename)
    writer = tf.python_io.TFRecordWriter(filename)
    #sess = tf.InteractiveSession()
    for index in range(num_examples):
        image_path = paths_to_images[index]
        file_contents = tf.read_file(image_path)
        if parent_type=='jpeg':
            this_image = tf.image.decode_jpeg(file_contents).eval(session=sess)
            image_type = 'image_jpeg'
        elif parent_type=='png':
            this_image = tf.image.decode_png(file_contents).eval(session=sess)
            image_type = 'image_png'
        else:
            raise TypeError("Image type unsupported.")

        rows = this_image.shape[0]
        cols = this_image.shape[1]
        channels = this_image.shape[2]

        example = tf.train.Example(features=tf.train.Features(feature={
            'height': int64Feature(rows),
            'width': int64Feature(cols),
            'channels': int64Feature(channels),
            'label': int64Feature(int(labels[index])),
            image_type: bytesFeature(file_contents.eval(session=sess))}))

        writer.write(example.SerializeToString())
    #sess.close()

def ReadTFRecord(filename_queue, parent_type='jpeg'):
    """
    ------------------------------------------------------------------------------------------
     TC
     Reads a TFRecord file decoding images and corresponding labels.
     !!! NB: This function is poorly named and will have to be revised at a later date. !!!

       filename_queue - a tf queue of labels

       returns: a single image and its label from the queue
    ------------------------------------------------------------------------------------------
    """

    reader = tf.TFRecordReader()

    if parent_type=='jpeg':
        image_type = 'image_jpeg'
    elif parent_type=='png':
        image_type = 'image_png'

    _, serialized_example = reader.read(filename_queue)
    features = tf.parse_single_example(
        serialized_example,
        features={image_type: tf.FixedLenFeature([], tf.string),
                  'label': tf.FixedLenFeature([], tf.int64)})

    if parent_type=='jpeg':
        image = tf.image.decode_jpeg(features['image_jpeg'])
    elif parent_type=='png':
        image = tf.image.decode_png(features['image_png'])


    label = features['label']

    return image, label

"""
========================================================================
 Helper functions for data pre-processing.  Mapping feature ranges into
 specific intervals (linear mappings and gaussianisation mappings)
========================================================================
"""
def GaussianiseColumn(data, columnIndex):
    """
    ----------------------------------------------------------------------
     This function determines a parameterisation of the ratio of
     integrals of tests data to the error function in order to map
     test data to a Gaussian-like form.

     The data passed to the function will be mapped to the new basis
     and the shape parameters for this column, returned to the calling
     function to facilitate inverse mapping/mapping of additional
     samples of data with this feature.
    ----------------------------------------------------------------------
    """
    print ("""
  def GaussianiseColumn(data, columnIndex): work in progress

  This function needs to be written.
  """)


    return




def GetFeatureMaxMin(data, columnIndex):
    """
    ----------------------------------------------------------------------
     Given a feature in some data find the maximum and minimum values of that
     feature.
    ---------------------------------------------------------------------
    """
    MinRange = float(1e10)
    MaxRange = float(-1e10)

    for i in range(len(data)):
        thisRow = data[i]
        if(thisRow[columnIndex] < MinRange):
            MinRange = thisRow[columnIndex]
        if(thisRow[columnIndex] > MaxRange):
            MaxRange = thisRow[columnIndex]

    if (MaxRange == MinRange):
        print ("Maximum and minmum range indices match: [", MinRange, ", ", MinRange, "]")
        return 1.0, 0.0

    return MaxRange, MinRange




def GetFeatureRange(data, columnIndex):
    """
    ----------------------------------------------------------------------
     Given a feature in some data find the maximum and minimum values of that
     feature.  The linear transformation a*x+b can be determined from the range
     as x / (max-min) - min / (max-min).  This function computes the scale and
     offset required to map some feature space dimension [min, max] onto [0, 1]
     by determining the parameters a and b. These are returned to the calling
     function.

     If the range, max-min, is found to be zero then a=1.0 and b = 0 is
     returned.

     The transformation from [min, max] onto [0, 1] is achieved by
     computing:
                             x' = ax + b
    ----------------------------------------------------------------------
    """
    MaxRange, MinRange = GetFeatureMaxMin(data, columnIndex)
    Normalisation      = 1.0 / (MaxRange - MinRange)
    Offset             = -1.0 * MinRange * Normalisation

    return Normalisation, Offset




def NormaliseColumn(data, newMin=0, newMax=1):
    """
    ----------------------------------------------------------------------
     Apply the transformation: x' = ax + b to a column of data.  If newMin
     and newMax are supplied then the mapping is to the following value
      x'' = (ax + b)*(newMax-newMin) + newMin; where
         x has the range [min, max] given by data
         x' has the range [0, 1]
         x'' has the range [newMin, newMax]
    ----------------------------------------------------------------------
    """
    if(newMax == newMin):
        print ("unable to normalise data to a distribution with zero range")
        return data

    # compute the normalisation range given this is a column, the feature
    # index is 0.
    a, b = GetFeatureRange(data, 0)

    # ax+b maps the data onto [0, 1], we want to map onto [newMin, newMax]
    # therefore we have thisRange*(ax+b) + newMin as the function to map
    # x = [min, max] to x'' = [newMin, newMax]
    thisRange  = newMax-newMin
    thisOffset = newMin

    newData = []
    for i in range(len(data)):
        thisRow = data[i]
        theNewRow = []
        theNewRow.append( thisRange*(thisRow[0] * a + b) + newMin )
        newData.append(theNewRow)

    return newData




def NormaliseDataNeg1To1(data):
    """
    ----------------------------------------------------------------------
     Apply the transformation: x' = ax + b to each column of data to map
     column ranges to [-1.0, 1.0] for test/training/evaluation purposes
    ----------------------------------------------------------------------
    """
    FirstExample = data[0]
    MinRange = []
    MaxRange = []
    for i in range(len(FirstExample)):
        MinRange.append( -1.0 )
        MaxRange.append( 1.0 )

    theNewData = NormaliseData(data, MinRange, MaxRange)

    return theNewData



def NormaliseData(data, MinRange=[], MaxRange=[]):
    """
    ----------------------------------------------------------------------
     Apply the transformation: x' = ax + b to each column of data.
     The column ranges are mapped to [0, 1] unless specified otherwise.
    ----------------------------------------------------------------------
    """
    # determine the number of features in the data
    N = len(data[0])

    a = []
    b = []

    # compute the normalisation ranges for all columns of data
    for i in range(N):
        thisA, thisB = GetFeatureRange(data, i)
        a.append(thisA)
        b.append(thisB)

    # ax+b maps the data onto [0, 1], if we want to map onto [newMin, newMax]
    # therefore we have thisRange*(ax+b) + newMin as the function to map
    # x = [min, max] to x'' = [newMin, newMax].  For now just map to [0, 1]
    # for all data.
    newData = []
    if(len(MinRange) == 0) and (len(MaxRange)==0):
        # scale normalisation to [0, 1] for all columns
        for i in range(len(data)):
            thisRow = data[i]
            theNewRow = []
            for j in range(len(thisRow)):
                theNewRow.append( thisRow[j] * a[j] + b[j] )
            newData.append(theNewRow)
    else:
        # scale normalisation according to ranges supplied by the calling function
        for i in range(len(data)):
            thisRow = data[i]
            theNewRow = []
            for j in range(len(thisRow)):
                thisMin = MinRange[j]
                thisMax = MaxRange[j]
                thisRange = thisMax-thisMin
                theNewRow.append( thisRange * (thisRow[j] * a[j] + b[j] ) + thisMin )
            newData.append(theNewRow)

    return newData

def SplitLabelledData(data, train_ratio=0.5):
    """
    TC
    Splits labelled data into test and train given a train/test ratio.
    
    Arguments:
        data          - the data stored as tuples
        train_ratio   - the ratio of the total data that should be used for training

    Returns:
         train, test  - the data split into two folds
    """
    tuples = []
    n_examples = len(data)
    for key in data:
        tuples.append(data[key])
    random.shuffle(tuples)
    train = tuples[0:int(n_examples * train_ratio)]
    test = tuples[int(n_examples * train_ratio):n_examples]
    return train, test    

def OneHotConversion(labels):
    """
    TC
    Converts labels into one hot vectors.
    
    Arguments:
         labels   - a list of labels

    Returns:
         one-hots - a list of one hot vectors
    """
    max = 0
    for label in labels:
        if int(label) > max:
            max = int(label)
    #print("largest label", max)
    one_hots =[]
    for label in labels:
        one_hot_label = [0] * (max + 1)
        #print(one_hot_label)
        one_hot_label[int(label)] = 1
        one_hots.append(one_hot_label)
    return one_hots

def BinaryClasses(labels, classA=0):
    """
    TC
    Changes multi-class labels to binary class labels.
    Uses optional argument classA to pick which label is preserved.
    All other labels are collapsed into a second class.

    Arguments:
         labels     - a list of multi-class labels
         classA     - the label to be preserved, all others collapsed

    Returns:
         new_labels - the binarised class labels
    """
    #changes labels like '0', '1', '2', '3'
    #outputs simple '0' and '1' where anyting not '0'
    #has been replaced by '1'
    #print("changing multi-class labels to binary classes")
    new_labels = []
    for label in labels:
        if label != classA:
            if classA:
                label = 1
            else:
                label = 0
        new_labels.append(label)
    return new_labels

def CrudeDownSample(image, ratio):
    """
    ------------------------------------------------------------------------------------
     TC
     Down samples an image by ratio, called crude because ratio must be 1/2 or smaller.

       image            - the image you wish to down sample
       ratio            - a number (float) less than 1/2

       Returns: new_image, the downsampled image
    -----------------------------------------------------------------------------------
    """
    # Error checking (this method is only sufficient for ratio <= 1/2)
    if ratio > float(1/2):
        raise ValueError("Unsupported ratio. Ratio must be less than 1/2.")
    # Get widths/heights of new/old images
    width = image.shape[0]
    height = image.shape[1]
    new_width = int(width * ratio)
    new_height = int(height * ratio)

    # Calculate kernel_size from ratio
    kernel_size = [int(1/ratio), int(1/ratio)]

    # Init new_image
    new_image = np.empty((new_width, new_height))

    # Loop over pixels in new_image
    for i in range(new_width):
        n1 = i * kernel_size[0]
        for j in range(new_height):
            n2 = j * kernel_size[1]
            new_pixel = 0
            # Loop over pixels in kernel, taking sum of values
            for k1 in range(kernel_size[0]):
                for k2 in range(kernel_size[1]):
                    new_pixel += image[n1 + k1, n2 + k2]
            # Average values
            new_pixel /= kernel_size[0] * kernel_size[1]
            new_image[i,j] = new_pixel

    return new_image



"""
========================================================================
 Drawing functions
========================================================================
"""

def ImGen(shape, greyscale=True):
    """
    ----------------------------------------------------------------------
     TC
     Generates a blank image of a specified shape.

        shape       - specifies the 2D shape of the image

        greyscale   - bool default=True, specifies the number of layers
                      (1 = greyscale, 3 = RGB, unsupported: 4 = RGBA)

        returns:
        null_image  - a numpy array of shape ((shape), 1) or ((shape), 3).
    ----------------------------------------------------------------------
    """
    null_image = np.zeros(shape)
    null_image.fill(255)
    if greyscale==True:
        null_image = np.expand_dims(null_image, 3)
    else:
        null_image = np.expand_dims(null_image, 3)
        null_image = np.concatenate((null_image,
                                     null_image,
                                     null_image),
                                    axis=2)
    return null_image

def InSquare(x, xc, y, yc, r):
    """
    ----------------------------------------------------------------------
     TC
     Performs a check on coordinates (x,y) to see if they are in a square
     parametrised by its centre (xc, yc) and edge length (r).

        x           - first coordinate to check
        y           - second coordinate to check
        xc          - first coordinate specifying the centre of the square
        yc          - second coordinate specifying the centre of the square
        r           - 1/2 edge length of the square

        returns:
        True if (x,y) lies inside the square else False.
    ----------------------------------------------------------------------
    """
    if (xc - r) < x < (xc + r) or (yc - r) < y < (yc + r):
        return True
    else:
        return False

def InCircle(x, xc, y, yc, r):
    """
    ----------------------------------------------------------------------
     TC
     Performs a check on coordinates (x,y) to see if they are in a circle
     parametrised by its centre (xc, yc) and radius (r).

        x           - first coordinate to check
        y           - second coordinate to check
        xc          - first coordinate specifying the centre of the circle
        yc          - second coordinate specifying the centre of the circle
        r           - radius of the circle

        returns:
        True if (x,y) lies inside the circle else False.
    ----------------------------------------------------------------------
    """
    circle = (x - xc) ** 2 + (y - yc) ** 2
    R = r ** 2
    if circle < R:
        return True
    else:
        return False

def InRing(x, xc, y, yc, rmax, rmin):
    """
    ----------------------------------------------------------------------
     TC
     Performs a check on coordinates (x,y) to see if they are in a ring
     parametrised by its centre (xc, yc), outer radius (rmax) and inner
     radius (rmin).

        x           - first coordinate to check
        y           - second coordinate to check
        xc          - first coordinate specifying the centre of the ring
        yc          - second coordinate specifying the centre of the ring
        rmax        - outer radius of the ring
        rmin        - inner radius of the ring

        returns:
        True if (x,y) lies inside the ring else False.
    ----------------------------------------------------------------------
    """
    if rmax < rmin:
        # correct erroneously entered min/max
        temp = rmax
        rmax = rmin
        rmin = temp
    ring = (x - xc) ** 2 + (y - yc) ** 2
    Rmax = rmax ** 2
    Rmin = rmin ** 2
    if Rmin < ring < Rmax:
        return True
    else:
        return False

def DrawCircles(image, centres, radii, intensities, ring=None, radii_min=None):
    """
    ----------------------------------------------------------------------
     TC
     Takes a several lists of parameters and draws the corresponding features
     onto the image.

        For the ith feature the lists are read as follows:

        centres[i]     -       string: "xc, yc" two coordinates separated by a comma
                                   that specify the centre of the feature

        radii[i]       -       float:  A number that specifies the radius of the feature
                                   if radii_min[i] is specified then radii acts like
                                   an outer radius

        intensities[i] -       string: "vmin, vmax" two numbers separated by a comma
                                     that specify the max and min pixel intensities

        ring[i]        -        bool: specifies whether or not feature is a ring
                                      if True radii_min[i] must be specified

        radii_min[i]   -      float: A number that specifies the inner radius of the feature

    ----------------------------------------------------------------------
    """
    # get the width and height of the image
    width = image.shape[0]
    height = image.shape[1]

    # perform a quick check to see that lists are of compatible size
    num_circles = len(centres)
    num_radii = len(radii)
    num_intensities = len(intensities)
    if not num_circles == num_radii == num_intensities:
        print("incompatible sized inputs")

    # loop over width then height then num_circles
    for x in range(width):
        for y in range(height):
            for i in range(num_circles):
                # for this circle in the loop get the essential parameters
                this_centre = centres[i].split(',')
                this_radius = radii[i]
                xc = float(this_centre[0])
                yc = float(this_centre[1])
                r = float(this_radius)

                # check if this (x,y) coordinate is in the square specified by xc, yc and r.
                if InSquare(x, xc, y, yc, r):
                    this_intensity = intensities[i].split(',')
                    grey_min = float(this_intensity[0])
                    grey_max = float(this_intensity[1])
                    if grey_max < grey_min:
                        temp = grey_max
                        grey_max = grey_min
                        grey_min = temp

                    # get ring option
                    if not ring:
                        outline_this = False
                    else:
                        outline_this = ring[i]

                    # if not a ring then check if (x,y) is in circle
                    if not outline_this:
                        if InCircle(x, xc, y, yc, r):
                            # if (x,y) is in circle, fill the pixel
                            image[x,y] = random.uniform(grey_min, grey_max)
                    else:
                        # do the same but for the ring
                        if not radii_min:
                            print("please specify radii_min: see code for more info")
                        if not len(radii_min) == num_circles:
                            print("wrong number of thicknesses specified")
                        else:
                            this_rmin = float(radii_min[i])
                        if InRing(x, xc, y, yc, r, this_rmin):
                            image[x,y] = random.uniform(grey_min, grey_max)



def NoiseGen(image, grey_min, grey_max):
    """
    ----------------------------------------------------------------------
     TC
     Generates uniform noise and applies it to an image.

       image            - the image you want to apply noise to.
       grey_min         - min intensity for noise
       grey_max         - max intensity for noise
    ----------------------------------------------------------------------
    """
    width = image.shape[0]
    height = image.shape[1]
    noise  = np.zeros((width, height))
    for x in range(width):
        for y in range(height):
            noise[x,y] = random.uniform(grey_min, grey_max)
    return noise





"""
========================================================================
 Miscellaneous functions
========================================================================
"""

def TestResult(flag, textstring):
    """
    ----------------------------------------------------------------------
     print out boolean status for test scripts
    ----------------------------------------------------------------------
    """
    if flag == 1:
        print (textstring, " OK")
    else:
        print (textstring, " FAILED")


def TensorBoardInit(tflogs="/tmp/tensorflowlogs"):
    """
    ----------------------------------------------------------------------
     Setup storage of the TF graph
    ----------------------------------------------------------------------
    """
    merged = tf.summary.merge_all()
    writer = tf.summary.FileWriter(tflogs, graph=tf.get_default_graph())

    return

def TensorBoardFinalise(tflogs):
    """
    ----------------------------------------------------------------------
     Remind the user of the location of tensorboard log files and provide
     the command line instruciton required to inspect the graph.
    ----------------------------------------------------------------------
    """
    print ("""

Tensorboard summary files have been written.  In order to inspect the
graph please make sure you run the following command on completion:

""")
    print ("    tensorboard --logdir=", tflogs)

    return

def CurrentTime():
    current_time = time.time()
    return current_time

def TimeInterval(initial):
    current_time = time.time()
    interval = current_time - initial
    return interval

def PickElement(_list):
    """
    ----------------------------------------------------------------------
     TC
     Picks a random element from a list

       _list - a list

       Returns: a random element from _list
    ----------------------------------------------------------------------
    """
    return random.choice(_list)

def IncrementDirectory(parent_name, dir_name, delimitter='-', num_digits=4):
    """
    ----------------------------------------------------------------------------------
    Given a parent_name and a dir_name, searches parent_name for directories
    containing dir_name. Looks for the highest number after the delimitter,
    returns full path to this directory name incremented by 1.
    i.e.
    parent_name/dir_name-0001       <-- finds 1
    returns: parent_name/dir_name-0002
    
    parent_name            - string: absolute path to parent directory
    dir_name               - string: name of dir you wish to increment
    
    returns:
    new_dir - absolute path to the incremented directory
    ----------------------------------------------------------------------------------
    N.B. you still need to create it if that is what you desire!
    """
    existing_dirs = hipster.DirectoryList(parent_name)
    #print("current directories:", existing_dirs)

    highest_number = 0
    if not len(existing_dirs) == 0:
        #highest_number = 1
        for i in range(len(existing_dirs)):
            if name in existing_dirs[i]:
                this_number = existing_dirs[i].split(delimitter)[-1:]
                #print("result from split", this_number)
                this_number =  this_number[0].lstrip("0")
                try:
                    this_number = int(this_number)
                    #print("result from int cast:", this_number)
                    if this_number > highest_number:
                        highest_number = this_number
                except:
                    continue

    n = str(highest_number + 1).zfill(num_digits)
    new_dir = parent_name + '/' + dir_name + delimitter + n
    return new_dir
