#!/usr/bin/python
"""
Example_MLP.py

Construct a 2 layer regression MLP to use as a function approximator.

This python script has been written for use with the PracticalMachineLearning
course used for the Queen Mary University of London Summer School.

Please see
   https://qmplus.qmul.ac.uk/course/view.php?id=10006
  ----------------------------------------------------------------------
  author:       Adrian Bevan (a.j.bevan@qmul.ac.uk)
  Copyright (C) QMUL 2018
  ----------------------------------------------------------------------
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Network training parameters
learning_rate    = 0.01
training_epochs  = 1000
min_x = -10
max_x = 10
Ngen  = 10000

# Network architecture parameters
n_input    = 1   # 1D function, so there is one input feature per example
n_classes  = 1   # Regression output is single valued
n_hidden_1 = 50 # 1st layer num features
n_hidden_2 = 50 # 2nd layer num features

print("--------------------------------------------")
print("Number of input features       = ", n_input)
print("Number of output classes       = ", n_classes)
print("Number of examples to generate = ", Ngen)
print("Learning rate            alpha = ", learning_rate)
print("Number of training epochs      = ", training_epochs)
print("--------------------------------------------")

def myFunctionTF(arg):
    """
    User defined function for the MLP to learn.  The default example is 
    the square root function.
    """
#    return tf.sqrt(arg)
    return tf.square(arg)
#    return arg+tf.square(arg)+arg*tf.square(arg)
#    return arg+tf.square(arg)+arg*tf.square(arg)+tf.square(arg)*tf.square(arg)
#    return tf.sin(arg)


# tf Graph input:
#  x: is the tensor for the input data (the placeholder entry None is used for that;
#     and the number of features input (n_input = 1).
#
#  y: is the tensor for the output value of the function that is being approximated by 
#     the MLP.
#
x = tf.placeholder(tf.float32, [None, n_input], name="x")
y = tf.placeholder(tf.float32, [None, n_classes], name="y")


# We construct layer 1 from a weight set, a bias set and the activiation function used
# to process the impulse set of features for a given example in order to produce a 
# predictive output for that example.
#
#  w_layer_1:    the weights for layer 1.  The first index is the input feature (pixel)
#                and the second index is the node index for the perceptron in the first
#                layer.
#  bias_layer_1: the biases for layer 1.  There is a single bias for each node in the 
#                layer.
#  layer_1:      the activation functions for layer 1
#
w_layer_1    = tf.Variable(tf.random_normal([n_input, n_hidden_1]))
bias_layer_1 = tf.Variable(tf.random_normal([n_hidden_1]))
layer_1      = tf.nn.relu(tf.add(tf.matmul(x,w_layer_1),bias_layer_1))

# We construct layer 2 in an analogous way; now we have a second set of weights,
# biases and activation functions.  These are named similarly to the first set.
# the changes are:
#   1) the input number of features to the second layer, corresponds to the number
#      of nodes in the first
#   2) layer 1 is used in the matrix multiplication when constructing layer 2 
# as opposed to the data placeholder x.
#
#  w_layer_2:    the weights for layer 2.  The first index is the input feature (pixel)
#                and the second index is the node index for the perceptron in the first
#                layer.
#  bias_layer_2: the biases for layer 2.  There is a single bias for each node in the 
#                layer.
#  layer_2:      the activation functions for layer 2
#
w_layer_2    = tf.Variable(tf.random_normal([n_hidden_1, n_hidden_2]))
bias_layer_2 = tf.Variable(tf.random_normal([n_hidden_2]))
layer_2      = tf.nn.relu(tf.add(tf.matmul(layer_1,w_layer_2),bias_layer_2))

# Similarly we now construct the output of the network, where the output layer
# combines the information down into a space of evidences for the 10 possible
# classes in the problem.
output       = tf.Variable(tf.random_normal([n_hidden_2, n_classes]))
bias_output  = tf.Variable(tf.random_normal([n_classes]))
output_layer = tf.matmul(layer_2, output) + bias_output

#optimise with l2 loss function
print("Using the L2 loss function implemented in tf.nn")
cost = tf.nn.l2_loss(output_layer - y)

# optimizer: take the Adam optimiser, see https://arxiv.org/pdf/1412.6980v8.pdf for
# details of this algorithm.
print("Using the Adam optimiser to train the network")
optimizer = tf.train.AdamOptimizer(learning_rate=learning_rate).minimize(cost)

# generate data, the input data is a random number betwen 1 and 10,
# and the corresponding label value is the square root of that number
tftraindata = tf.random_uniform([Ngen, 1], min_x, max_x)  # training set
tftestdata  = tf.random_uniform([Ngen, 1], min_x, max_x)  # test set

# define operation for computing the regression output
probabilities = output_layer

# Initializing the variables
init  = tf.global_variables_initializer()

# Start the session to embark on the training cycle
sess = tf.Session()
sess.run(init)

# convert the training data to np arrays so that these can be used with the feed_dict when training
traindata  = sess.run(tftraindata) 
target_value = sess.run(myFunctionTF(traindata))

# convert the test data to np arrays so that these can be used with the feed_dict when training
testdata  = sess.run(tftestdata) 
test_value = sess.run(myFunctionTF(testdata))

# plot the test and training data
plotting = 1
if plotting:
    # plot the function f(x)=sqrt(x)
    plt.subplot(2, 1, 1)
    plt.plot(testdata, test_value, 'r.')
    plt.ylabel('f(x) = sqrt(x)')
    plt.xlabel('x')
    plt.title('Training data')
    plt.subplot(2, 1, 2)
    plt.plot(traindata, target_value, 'b.')
    plt.ylabel('f(x) = sqrt(x)')
    plt.xlabel('x')
    plt.title('Test data')
    plt.show()

# arrays to compare the input value vs the prediction
input_value      = []
epoch_set        = []
cost_test_value  = []
cost_set         = []
prediction_value = []

# Training cycle - for this example we do not batch train as this is a shallow network with 
# a low number of training events.
for epoch in range(training_epochs):
    the_cost = 0.
    
    print("Training epoch number ", epoch)
    sess.run(optimizer, feed_dict={x: traindata, y: target_value})
    the_cost = sess.run(cost, feed_dict={x: traindata, y: target_value})

    cost_set.append(the_cost)
    epoch_set.append(epoch+1)
    
    # Test model - this is a classification problem speciifc bit of code - irrelevant for 
    # a regression analysis
    #correct_prediction = tf.equal(tf.argmax(output_layer, 1), tf.argmax(y, 1))
    #accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))
    #print ("Model Accuracy:", accuracy.eval( session=sess, feed_dict={x: traindata, y: target_value} ))

    the_cost = sess.run(cost, feed_dict={x: testdata, y: test_value})
    cost_test_value.append(the_cost)

    #
    # This is a regression analysis problem, so we want to evaluate and display the output_layer
    # value (model response function), and not an output prediction (which would have been appropraite
    # for a classification problem)
    #
    if epoch == training_epochs-1:
        step = (max_x - min_x)/100
        for i in range(100):
            thisx = min_x + i*step
            pred = probabilities.eval(feed_dict={x: [[thisx]]}, session=sess)
            print ("x = ", thisx, ", prediction =", pred)
            input_value.append(thisx)
            prediction_value.append(pred[0])

            pred = probabilities.eval(feed_dict={x: [[-thisx]]}, session=sess)
            print ("x = ", -thisx, ", prediction =", pred)
            input_value.append(-thisx)
            prediction_value.append(pred[0])


# Result summary
print ("Training phase finished")

plt.subplot(2, 1, 1)
plt.plot(testdata, test_value, 'b.')
plt.ylabel('f(x) = sqrt(x)')
plt.xlabel('x')
plt.title('Test data')
plt.plot(input_value, prediction_value, 'r*')
plt.ylabel('f(x) = sqrt(x)')
plt.xlabel('x')
plt.title('Network Response Function')

plt.subplot(2, 1, 2)
plt.plot(epoch_set, cost_set, 'o', label='MLP Training phase')
plt.plot(epoch_set, cost_test_value, 'rx', label='MLP Training phase')
plt.ylabel('cost')
plt.xlabel('epoch')

plt.show()

